(***************************************************************************

Abstract syntax trees for JavaCard source files

VerifiCard Project - Dmons research team - LRI - Universit Paris XI

$Id: ast.ml,v 1.21 2003/10/21 14:25:50 marche Exp $

***************************************************************************)

(* be careful, all printing functions functions use [Format] module *)

open Format;;
open Format_misc;;
open Ast_types;;
open Java_context;;


(*s Qualified identifiers *)

let rec fprintf_identifier form (loc,id) =
  fprintf form "%s" id

let rec fprintf_qualified_ident form qi =
  match qi with
    | [] -> fprintf form "(null ident ???)";
    | [i] -> fprintf_identifier form i
    | i::qi -> 
	fprintf_qualified_ident form qi;
	fprintf form ".";
	fprintf_identifier form i
;;

(*s Modifiers *)
 
let string_of_modifier (m : modifier) = 
  match m with
  | `STATIC -> "static"
  | `FINAL -> "final" 
  | `PUBLIC -> "public" 
  | `PRIVATE -> "private" 
  | `PROTECTED -> "protected" 
  | `NATIVE -> "native" 
  | `SYNCHRONIZED -> "synchronized" 
  | `ABSTRACT -> "abstract" 
  | `TRANSIENT -> "transient" 
      (* "threadsafe" ? *)
  | `PURE -> "pure" 
  | `SPEC_PUBLIC -> "spec_public" 
  | `MODEL -> "model"
;;  
    
let fprintf_modifier form m =
  fprintf form "%s" (string_of_modifier m)
;;

let fprintf_modifiers form mods =
  fprintf form "Modifiers@,(%a)" (fprintf_list fprintf_modifier)  mods
;;

  
(* Javadoc comments *)

let fprintf_javadoc form s =
  fprintf form "\"%s\"" s 
;;

(*s type expressions *)

let rec fprintf_type_expr form t =
  match t with
    | Base_type(s) -> fprintf form "Base_type(%s)" s
    | Type_name(qi) ->
	fprintf form "@[<hv 1>Type_name@,(";
	fprintf_qualified_ident form qi;
	fprintf form ")@]"
    | Array_type_expr(t) ->
	fprintf form "@[<hv 1>Array_type@,(";
	fprintf_type_expr form t;
	fprintf form ")@]"


(*s literals *)

let fprint_literal form l =
  match l with
  | Int(n) -> fprintf form "Int(%d)" n
  | Float(f) -> fprintf form "Float(%g)" f
  | Bool(b) -> fprintf form "Bool(%b)" b
  | String(s) -> fprintf form "String(\"%s\")" s
  | Char(c) -> fprintf form "Char('%c')" c
  | Null -> fprintf form "Null" 
;;

(* variable ids *)

let rec fprintf_variable_id form id =
    match id with
  | Simple_id(loc,s) ->
      fprintf form "Simple_id(%s)" s;
  | Array_id(i) ->
      fprintf form "@[<hv 1>Array_id@,(";
      fprintf_variable_id form i;
      fprintf form ")@]"
;;

(*s expressions *)

let fprintf_quantifier form q =
  match q with
    | Forall -> fprintf form "Forall"
    | Exists -> fprintf form "Exists"

let mk_expr loc e =
  { expr_loc = loc ;
    expr_node = e }
;;

let mk_lit loc l = mk_expr loc (Lit l);;

let mk_un loc op e = mk_expr loc (Un(op,e));;

let mk_prepost loc op e = mk_expr loc (PrePostIncrDecr(op,e));;

let mk_bin loc e1 op e2 = mk_expr loc (Bin(e1,op,e2));;

let mk_expr_no_loc = mk_expr Location.dummy ;;

let expr_true = mk_expr_no_loc (Lit(Bool(true)))
let expr_false = mk_expr_no_loc (Lit(Bool(false)))
let expr_zero = mk_expr_no_loc (Lit(Int(0)))

let fprintf_prepostincrdecr form op =
  fprintf form "%s"
    (match op with
       | `PREINCR -> "pre++"
       | `POSTINCR -> "post++"
       | `PREDECR -> "pre--"
       | `POSTDECR -> "post--")

let rec expand_name l =
  match l with
    | [] -> assert false
    | [(loc,_) as id] -> mk_expr loc (Var(id)) 
    | ((loc,_) as id)::l -> 
	let e = expand_name l in
	let new_loc = Location.merge_locs e.expr_loc loc in
	mk_expr new_loc
	  (Field_access(Primary_access(e,id))) 

let rec build_array_type t n =
  if n=0 then t else Array_type_expr(build_array_type t (pred n))
    
let rec build_array_creation_expr t (l,n) =
  match l with
    | [] -> 
	Implicit_array_creation(build_array_type t n)
    | a::b ->
	Explicit_array_creation(a,(build_array_creation_expr t (b,n)))
	  
let rec fprintf_field_access form fa =
  match fa with
    | Super_access(loc,id) -> fprintf form "Super_access(%s)" id
    | Primary_access(e,id) ->
	fprintf form "@[Primary_access@,(";
	fprintf_expr form e;
	fprintf form ",@ %a)@]" fprintf_identifier id
    | Typed_primary_access(e,ive) ->
	fprintf form "@[Typed_primary_access@,(%a,@ %a)@]"
	  fprintf_expr e
	  Java_context.fprintf_instance_variable_entry ive
    | Typed_constant_access(c) ->
	fprintf form "@[Typed_constant_access@,(%a)@]"
	  Java_context.fprintf_constant_entry c
    | Array_length(e) ->
	fprintf form "@[<hv 2>Array_length@,(%a)@]"
	  fprintf_expr e 

and fprintf_expr form e =
  match e.expr_node with 
    | Lit(l) -> 
	fprintf form "@[<hv 2>Lit@,(";
	fprint_literal form l;
	fprintf form ")@]"
    | Bin(e1,op,e2) ->
	fprintf form "@[<hv 2>Bin@,(";
	fprintf_expr form e1;
	fprintf form ",@ %s,@ " op;
	fprintf_expr form e2;
	fprintf form ")@]"
    | Un(op,e) ->
	fprintf form "@[<hv 2>Un(%s," op;
	fprintf_expr form e;
	fprintf form ")@]"
    | PrePostIncrDecr(op,e) ->
	fprintf form "@[<hv 2>PrePostIncrDecr(%a,%a)@]" 
	fprintf_prepostincrdecr op
	fprintf_expr e
    | Var(l,x) -> fprintf form "Var(%s)" x
    | Typed_var(ve) -> 
	fprintf form "Typed_var(%a)" Java_context.fprintf_local_variable_entry ve
    | Static_class(ce) -> 
	fprintf form "Static_class(%a)" Java_context.fprintf_class_entry ce
    | Static_interface(ie) -> 
	fprintf form "Static_interface(%a)" Java_context.fprintf_interface_entry ie
    | This -> fprintf form "This"
    | Assign_name(x,op,e) ->
	fprintf form "@[<hv 2>Assign_name(";
	fprintf_qualified_ident form x;
	fprintf form ",@ %s,@ " op;
	fprintf_expr form e;
	fprintf form ")@]"
    | Assign_local_var(x,op,e) ->
	fprintf form "@[<hv 2>Assign_local_var(";
	Java_context.fprintf_local_variable_entry form x;
	fprintf form ",@ %s,@ " op;
	fprintf_expr form e;
	fprintf form ")@]"
    | Assign_field(f,id,e) ->  
	fprintf form "@[<hv 2>Assign_field(";
	fprintf_field_access form f;
	fprintf form ",@ %s,@ " id;
	fprintf_expr form e;
	fprintf form ")@]"
    | Assign_array(e1,e2,op,e3) ->
	fprintf form "@[<hv 2>Assign_array(";
	fprintf_expr form e1;
	fprintf form ",@ ";
	fprintf_expr form e2;
	fprintf form ",@ %s,@ " op;
	fprintf_expr form e3;
	fprintf form ")@]"
    | Typed_assign_array(t,e1,e2,op,e3) ->
	fprintf form "@[<hv 2>Typed_assign_array(%a,@ "
	  Java_context.fprintf_java_type t;
	fprintf_expr form e1;
	fprintf form ",@ ";
	fprintf_expr form e2;
	fprintf form ",@ %s,@ " op;
	fprintf_expr form e3;
	fprintf form ")@]"
    | If_expr(e1,e2,e3) ->
	fprintf form "@[<hv 2>If_expr@,(";
	fprintf_expr form e1;
	fprintf form ",";
	fprintf_expr form e2;
	fprintf form ",";
	fprintf_expr form e3;
	fprintf form ")@]"
    | Field_access(fa) ->
	fprintf form "@[<hv 2>Field_access@,(";
	fprintf_field_access form fa;
	fprintf form ")@]"
    | Method_call(e,id,args) ->
	fprintf form "@[<hv 2>Method_call@,(";
	fprintf_option fprintf_expr form e;
	fprintf form ",@ %a,@ %a)@]" fprintf_identifier id
	  (fprintf_list fprintf_expr) args
    | Typed_method_call(e,me,args) ->
	fprintf form "@[<hv 2>Typed_method_call@,(";
	fprintf_expr form e;
	fprintf form ",@ %a,@ %a)@]" fprintf_method_entry me
	  (fprintf_list fprintf_expr) args
    | Array_creation(a) ->
	fprintf form "@[<hv 2>Array_creation@,(";
	fprintf_array_creation form a;
	fprintf form ")@]"
    | Array_access(e1,e2) ->
	fprintf form "@[<hv 2>Array_access@,(%a,@ %a)@]"
	  fprintf_expr e1 fprintf_expr e2 
    | Typed_array_access(t,e1,e2) ->
	fprintf form "@[<hv 2>Typed_array_access@,(%a,@ %a,@ %a)@]"
	  Java_context.fprintf_java_type t fprintf_expr e1 fprintf_expr e2 
    | Cast(t,e) ->
	fprintf form "@[<hv 2>Cast@,(";
	fprintf_type_expr form t;
	fprintf form ",";
	fprintf_expr form e;
	fprintf form ")@]"
    | Typed_cast(t,e) ->
	fprintf form "@[<hv 2>Typed_cast@,@[<hv 1>(%a,@,%a)@]@]"
	Java_context.fprintf_java_type t
	fprintf_expr  e
    | Class_instance_creation(id,args) ->
	fprintf form "@[<hv 2>Class_instance_creation@,(";
	fprintf_qualified_ident form id;
	fprintf form ", %a)@]"
	  (fprintf_list fprintf_expr) args
    | Typed_class_instance_creation(ce,args) ->
	fprintf form "@[<hv 2>Typed_class_instance_creation@,(";
	Java_context.fprintf_constructor_entry form ce;
	fprintf form ", %a)@]"
	  (fprintf_list fprintf_expr) args
    | Instanceof(e,t) ->
	fprintf form "@[<hv 2>Instanceof@,(%a,@ %a)@]" 
	  fprintf_expr e fprintf_type_expr t
    | Typed_instanceof(e,t) ->
	fprintf form "@[<hv 2>Typed_instanceof@,(%a,@ %a)@]" 
	  fprintf_expr e Java_context.fprintf_java_type t
    | Result -> fprintf form "Result"
    | Old(e) ->
	fprintf form "@[<hv 2>Old@,(%a)@]" 
	  fprintf_expr e 
    | Fresh(e) ->
	fprintf form "@[<hv 2>Fresh@,(%a)@]" 
	  fprintf_expr e 
    | Quantifier(q,t,idl,e1,e2) ->
	fprintf form "@[<hv 2>Quantifier@,(%a,%a,%a,%a%a)@]" 
	  fprintf_quantifier q
	  fprintf_type_expr t 
	  (fprintf_list fprintf_variable_id) idl
	  fprintf_expr e1 
	  fprintf_expr e2 
    | Typed_quantifier(q,idl,e1,e2) ->
	fprintf form "@[<hv 2>Typed_quantifier@,(%a,%a,%a%a)@]" 
	  fprintf_quantifier q
	  (fprintf_list Java_context.fprintf_local_variable_entry) idl
	  fprintf_expr e1 
	  fprintf_expr e2 




and fprintf_array_creation form a =
  match a with
    | Explicit_array_creation(e,a) ->
	fprintf form "Explicit_array_creation@,(";
	fprintf_expr form e;
	fprintf form ",";
	fprintf_array_creation form a;
	fprintf form ")"
    | Implicit_array_creation(t) ->
	fprintf form "Implicit_array_creation@,(";
	fprintf_type_expr form t;
	fprintf form ")"
    | Typed_implicit_array_creation(t) ->
	fprintf form "Typed_implicit_array_creation@,(";
	Java_context.fprintf_java_type form t;
	fprintf form ")"

let printf_expr e = fprintf_expr std_formatter e

(*s variable declarations *)
      

let rec fprintf_variable_initializer form i =
  match i with
  | Simple_initializer(e) ->
      fprintf form "@[<hv 1>Simple_initializer@,(";
      fprintf_expr form e;
      fprintf form ")@]"
  | Array_initializer(l) ->
      fprintf form "@[<hv 1>Array_initializer@,(%a)@]"
	(fprintf_list fprintf_variable_initializer) l
;;

 
      
let fprintf_variable_declarator form v =
  fprintf form "@[<hv 2>{ @[<hv 2>variable_id =@ ";
  fprintf_variable_id form v.variable_id;
  fprintf form " ;@]@ @[<hv 2>variable_initializer =@ %a"
    (fprintf_option fprintf_variable_initializer) v.variable_initializer;
  fprintf form "@]}@]"


let fprintf_variable_declaration form v =
  fprintf form "@[<hv 2>{ @[<hv 2>variable_modifiers =@ ";
  fprintf_modifiers form v.variable_modifiers;
  fprintf form " ;@]@ @[<hv 2>variable_type =@ ";
  fprintf_type_expr form v.variable_type;
  fprintf form " ;@]@ @[<hv 2>variable_decls =@ %a@]}@]"
    (fprintf_list fprintf_variable_declarator) v.variable_decls
;;




(*s formal parameters *)

let rec fprintf_parameter form t =
  match t with
  | Simple_parameter(t,s) ->
      fprintf form "@[<hv 1>Simple_parameter@,(";
	fprintf_type_expr form t;
	fprintf form ",@ %a)@]" fprintf_identifier s
  | Array_parameter(p) ->
	fprintf form "@[<hv 1>Array_parameter@,(";
	fprintf_parameter form p;
	fprintf form ")@]"
;;
  
(*s JML annotations *)

let fprintf_privacy form p =
  match p with
  | Privacy_none -> fprintf form "None"
  | Privacy_public  -> fprintf form "Public"
  | Privacy_protected  -> fprintf form "Protected"
  | Privacy_private -> fprintf form "Private"
;;


let fprintf_assignable_object form (obj,e) =
  match obj with
    | Assignable_everything -> fprintf form "Everything"
    | Assignable_nothing -> fprintf form "Nothing"
    | Assignable_set_ref(e,_) -> fprintf form "Expr(%a,...)" fprintf_expr e
    | Typed_assignable_set_ref(t,e,_) -> 
	fprintf form "Typed_expr(%a,%a,...)" 
	  Java_context.fprintf_java_type t
	  fprintf_expr e
    | Assignable_all_fields(e) -> 
	fprintf form "AllFields(%a)" 
	fprintf_expr e
;;

let fprintf_assignable_clause form l =
  fprintf form "@[<hv 2>Assignable(%a)@]"
    (fprintf_list fprintf_assignable_object) l
;;

let default_loop_annotation =
  { loop_invariant = expr_true ; 
    loop_modifies = None;
    typed_loop_modifies = None;
    loop_decreases = expr_zero}
;;

let fprintf_loop_annotation form l =
  fprintf form "@[<hv 2>{ loop_invariant = %a;@ decreases = %a }@]"
    fprintf_expr l.loop_invariant
    fprintf_expr l.loop_decreases
;;



let default_spec_header =
  { requires_clauses = [] ;
    when_clauses = [] ;
    measured_clauses = [] }
;;
 

let fprintf_normal_spec_case form spec =
  fprintf form "<...normal spec...>"



let fprintf_spec_case form spec =
  match spec with
  | Behavior(_,_) ->
      fprintf form 
	"@[<hv 2>Behaviour@,(...)@]"
  | Normal_behavior(_,nsc) ->
      fprintf form 
      "@[<hv 2>Normal_behaviour@,(...,%a)@]"
        fprintf_normal_spec_case nsc
  | Exceptional_behavior(_,_) ->
      fprintf form 
	"@[<hv 2>Exceptional_behaviour@,(...)@]"
  | Generic _ ->
      fprintf form 
	"@[<hv 2>Geneirc@,(...)@]"
;;


let default_method_specification =
  Non_extending([])

exception Not_a_statement_spec of string;;

let to_statement_spec spec =
  match spec with
  | Non_extending scl ->
      begin
	match scl with
	  | [Generic s] -> 
	      begin
	      if s.generic_spec_var_decls.forall_var_decls <> [] 
		or s.generic_spec_var_decls.let_var_decls <> [] 
	      then raise (Not_a_statement_spec "no var decl allowed here");
	      if s.generic_spec_header.when_clauses <> [] 
	      then raise (Not_a_statement_spec "when clause not supported here");
	      if s.generic_spec_header.measured_clauses <> [] 
	      then raise (Not_a_statement_spec "measured_by clause not supported here");
	      match s.generic_spec_body with
		| Simple_spec_body sb ->
		    if sb.diverges_clauses <> []
		    then raise (Not_a_statement_spec "diverges clause not supported here");
		    {
		      statement_requires = s.generic_spec_header.requires_clauses;
		      statement_assignable = sb.assignable_clauses;
		      statement_ensures = sb.ensures_clauses;
		      statement_signals = sb.signals_clauses;
		    }
		| Nested_spec_body _ ->
		    raise (Not_a_statement_spec "nested spec not allowed here");
	      end
	  | _ -> raise (Not_a_statement_spec "only generic spec allowed here")
      end
  | _ -> raise (Not_a_statement_spec "only non-extending spec allowed here")
      

let concat_specs s1 s2 =
  match (s1,s2) with
  | Non_extending(l1),Extending_also(l2) ->
      Non_extending(l1@l2)
  | _ -> assert false
;;

let fprintf_method_specification form spec =
  match spec with
  | Non_extending scl ->
      fprintf form 
	"@[<hv 2>Non_extending@,(%a)@]"
	(fprintf_list fprintf_spec_case) scl
  | Compact _ ->
      fprintf form 
	"@[<hv 2>Compact@,(...)@]"
  | Typed_compact _ ->
      fprintf form 
	"@[<hv 2>Typed_compact@,(...)@]"
  | Extending_also _ ->
      fprintf form 
	"@[<hv 2>Extending_also@,(...)@]"
  | Extending_and _ ->
      fprintf form 
	"@[<hv 2>Extending_and@,(...)@]"
;;


let fprintf_jml_variable_declaration form v =
  fprintf form "@[<hv 2>{ @[<hv 2>jml_variable_modifiers =@ ";
  fprintf_modifiers form v.jml_variable_modifiers;
  fprintf form " ;@]@ @[<hv 2>jml_variable_type =@ ";
  fprintf_type_expr form v.jml_variable_type;
  fprintf form " ;@]@ @[<hv 2>jml_variable_decls =@ %a@]}@]"
    (fprintf_list fprintf_variable_id) v.jml_variable_decls
;;


let default_jml_declaration = Invariant(expr_true);;

let fprintf_jml_declaration form d =
  match d with
  | Invariant(e) ->
      fprintf form 
	"@[<hv 2>Invariant@,(%a)@]" fprintf_expr e
  | Jml_variable_declaration(d) ->
      fprintf form 
	"@[<hv 2>Jml_variable_declaration@,(%a)@]" 
	fprintf_jml_variable_declaration d
  | Represents(id,e) ->
      fprintf form 
	"@[<hv 2>Represents@,(%a,%a)@]" 
	fprintf_identifier id
	fprintf_expr e
  | Typed_represents(ive,e) ->
      fprintf form 
	"@[<hv 2>Typed_represents@,(%a,%a)@]" 
	Java_context.fprintf_instance_variable_entry ive
	fprintf_expr e
;;


let default_jml_specification = Jml_declaration([]);;

(*s statements *)


let make_block l =
  match l with
    | [] -> Skip
    | [x] -> x
    | _ -> Block(l)
;;

let fprintf_switch_label form l =
  match l with
    | Case(e) ->
	fprintf form "@[<hv 1>Case@,(";
	fprintf_expr form e;
	fprintf form ")@]"
    | Default -> fprintf form "Default"
		     
let rec fprintf_statement form s =
  match s with
    | Skip  -> fprintf form "Skip"
    | Expr(e) ->
	fprintf form "@[<hv 1>Expr@,(";
	fprintf_expr form e;
	fprintf form ")@]"
    | Typed_expr(t,e) ->
	fprintf form "@[<hv 1>Typed_expr@,(%a,@,%a)@]"
	  Java_context.fprintf_java_type t
	  fprintf_expr e
    | Var_decl(v) ->
	fprintf form "@[<hv 1>Var_decl@,(";
	fprintf_variable_declaration form v;
	fprintf form ")@]"
    | Typed_var_decl(v,init,b) ->
	fprintf form "@[<hv 1>Typed_var_decl@,(";
	Java_context.fprintf_local_variable_entry form v;
	fprintf form ",@ %a"
	  (fprintf_option fprintf_variable_initializer) init;
	fprintf form ",@ ";
	fprintf_statement form b;
	fprintf form ")@]"
    | Return(e) ->
	fprintf form "@[<hv 1>Return@,(%a)@]"
	  (fprintf_option fprintf_expr) e
    | Break(l) ->
	fprintf form "Break(%s)" l
    | Continue(l) ->
	fprintf form "Continue(%s)" l
    | Label(l,s) ->
	fprintf form "@[<hv 1>Label@,(%s,@ " l;
	fprintf_statement form s;
	fprintf form ")@]"
    | If_statement(e,s1,s2) ->
	fprintf form "@[<hv 1>If@,(";
	fprintf_expr form e;
	fprintf form ",@ ";
	fprintf_statement form s1;
	fprintf form ",@ ";
	fprintf_statement form s2;
	fprintf form ")@]"
    | While(loop_inv,e,s) ->
	fprintf form "@[<hv 1>While@,(%a,@ "
	  fprintf_loop_annotation loop_inv;	
	fprintf_expr form e;
	fprintf form ",@ ";
	fprintf_statement form s;
	fprintf form ")@]"
    | For(loop_inv,el1,e,el2,s) ->
	fprintf form "@[<hv 1>For@,(%a,@ %a,@ %a,@ %a,@ %a)@]"
	  fprintf_loop_annotation loop_inv
	  (fprintf_list fprintf_statement) el1
	  fprintf_expr e
	  (fprintf_list fprintf_statement) el2
	  fprintf_statement s
    | For_decl(loop_inv,decl,e,el2,s) ->
	fprintf form "@[<hv 1>For_decl@,(%a,@ %a,@ %a,@ %a,@ %a)@]"
	  fprintf_loop_annotation loop_inv
	  fprintf_variable_declaration decl
	  fprintf_expr e
	  (fprintf_list fprintf_statement) el2
	  fprintf_statement s
    | For_typed_decl(lve,lvi,loop_inv,e,el2,s) ->
	fprintf form "@[<hv 1>For_typed_decl@,(%a,@ %a,@ %a,@ %a,@ %a,@ %a)@]"
	  fprintf_local_variable_entry lve
	  (fprintf_option fprintf_variable_initializer) lvi
	  fprintf_loop_annotation loop_inv
	  fprintf_expr e
	  (fprintf_list fprintf_statement) el2
	  fprintf_statement s
    | Switch(e,cases) ->
	fprintf form "@[<hv 1>Switch@,(";
	fprintf_expr form e;
	fprintf form ",@ %a)@]"
	  (fprintf_list fprintf_case) cases
    | Block(s) ->
	fprintf form "@[<hv 1>Block@,(";
	fprintf_block form s;
	fprintf form ")@]"
    | Synchronized (_, _) -> assert false
    | Try (_, _, _) -> assert false
    | Do (_,_, _) -> assert false
    | Throw (_) -> assert false
    | Jml_assert(e) ->
	fprintf form "@[<hv 1>Jml_assert@,(%a)@]" fprintf_expr e
    | Annotated(spec,st) ->
	fprintf form "@[<hv 1>Annotated@,(...,%a)@]" fprintf_statement st
    | Typed_annotated(spec,st) ->
	fprintf form "@[<hv 1>Typed_annotated@,(...,%a)@]" fprintf_statement st

and fprintf_case form (case_list,bloc) =
  fprintf form "@[(";
  fprintf_list fprintf_switch_label form case_list;
  fprintf form ",@ ";
  fprintf_block form bloc;
  fprintf form ")@]"


and fprintf_block form s =
  fprintf_list fprintf_statement form s
;;

let printf_statement s = fprintf_statement std_formatter s

(*s method declarations *)

let rec fprintf_method_declarator form t =
  match t with
  | Simple_method_declarator(id,params) ->
      fprintf form "@[<hv 1>Simple_method_declarator(%a,@ " 
	fprintf_identifier id;
      fprintf_list fprintf_parameter form params;
      fprintf form ")@]"
  | Array_method_declarator(p) ->
	fprintf form "@[<hv 1>Array_method_declarator@,(";
	fprintf_method_declarator form p;
	fprintf form ")@]"
;;
  


let fprintf_method_declaration form m =
  fprintf form "@[<hv 2>{ @[<hv 2>method_modifiers =@ ";
  fprintf_modifiers form m.method_modifiers;
  fprintf form " ;@]@ @[<hv 2>method_return_type =@ ";
  fprintf_option fprintf_type_expr form m.method_return_type;
  fprintf form " ;@]@ @[<hv 2>method_declarator =@ ";
  fprintf_method_declarator form m.method_declarator;
  fprintf form " ;@]@ @[<hv 2>method_throws =@ ";
  fprintf_list fprintf_qualified_ident form m.method_throws;
  fprintf form "@]}@]"
;;


(*s constructor declarations *)



let fprintf_explicit_constructor_invocation form e =
  match e with
  | Invoke_none ->
      fprintf form "Invoke_none"
  | Invoke_this(l) ->
      fprintf form "@[<hv 2>Invoke_this@,(";
      fprintf_list fprintf_expr form l;
      fprintf form ")@]"
  | Invoke_super(l) ->
      fprintf form "@[<hv 2>Invoke_super@,(";
      fprintf_list fprintf_expr form l;
      fprintf form ")@]"
;;

      
let fprintf_constructor_declaration form c =
  fprintf form "@[<hv 2>{ @[<hv 2>constr_modifiers =@ ";
  fprintf_modifiers form c.constr_modifiers;
  fprintf form " ;@]@ @[<hv 2>constr_name =@ %a" 
    fprintf_identifier c.constr_name;
  fprintf form " ;@]@ @[<hv 2>constr_parameters =@ ";
  fprintf_list fprintf_parameter form c.constr_parameters;
  fprintf form " ;@]@ @[<hv 2>constr_throws =@ ";
  fprintf_list fprintf_qualified_ident form c.constr_throws;
  fprintf form " ;@]@ @[<hv 2>constr_body =@ (";
  match c.constr_body with
    | (inv,bloc) -> 
	fprintf_explicit_constructor_invocation form inv;
	fprintf form ",@ ";
	fprintf_block form bloc;
	fprintf form ")@]}@]"
;;



(*s class declarations *)
      


let fprintf_field_declaration form f =
  match f with
    | Method(m,b) -> 
	fprintf form "@[<hv 1>Method@,(";
	fprintf_method_declaration form m;
	fprintf form ",@ %a)@]"
	  (fprintf_option fprintf_block) b
    | Typed_method(m,jml,b) -> 
	fprintf form "@[<hv 1>Typed_method@,(";
	Java_context.fprintf_method_entry form m;
	fprintf form ",%a@ ,@ %a)@]"
	  fprintf_method_specification jml
	  (fprintf_option fprintf_block) b
    | Constructor(c) -> 
	fprintf form "@[<hv 1>Constructor@,(";
	fprintf_constructor_declaration form c;
	fprintf form ")@]"
    | Typed_constructor(c,jml,eci,b) -> 
	fprintf form "@[<hv 1>Typed_constructor@,(";
	Java_context.fprintf_constructor_entry form c;
	fprintf form ",<jml>,@ <...>,@  ";
	fprintf_block form b;
	fprintf form ")@]"
    | Variable(v) ->
	fprintf form "@[<hv 1>Variable@,(";
	fprintf_variable_declaration form v;
	fprintf form ")@]"
    | Typed_variable(v,init) ->
	fprintf form "@[<hv 1>Typed_variable@,(%a,@ %a)@]"
	  Java_context.fprintf_instance_variable_entry v
	  (Format_misc.fprintf_option fprintf_variable_initializer) init
    | Static_initializer(b) ->
	fprintf form "@[<hv 1>Static_initializer@,(%a)@]"
	fprintf_block b
    | Specification(s) ->
	fprintf form "@[<hv 1>Specification@,(%a)@]"
	  (fprintf_list fprintf_jml_declaration) s
  



  
let fprintf_commented_field_declaration form (doc,decl) =
  fprintf form 
    "@[<hv 2>{ doc_comment =@ %a;@ decl =@ %a;@ }@]" 
    fprintf_javadoc doc
    fprintf_field_declaration decl
;;


let fprintf_class_declaration form d =
  fprintf form "@[<hv 2>{ @[<hv 2>class_modifiers =@ ";
  fprintf_modifiers form d.class_modifiers;
  fprintf form " ;@]@ @[<hv 2>class_name =@ %a" 
    fprintf_identifier d.class_name;
  fprintf form " ;@]@ @[<hv 2>class_extends =@ ";
  fprintf_option fprintf_qualified_ident form d.class_extends;
  fprintf form " ;@]@ @[<hv 2>class_implements =@ ";
  fprintf_list fprintf_qualified_ident form d.class_implements;
  fprintf form " ;@]@ @[<hv 2>class_fields =@ ";
  fprintf_list fprintf_commented_field_declaration form d.class_fields;
  fprintf form "@]}@ @]"
;;

(*s interface declarations *)


let fprintf_interface_member_declaration form f =
  match f with
    | Method_header(m) -> 
	fprintf form "@[<hv 1>Method_header@,(";
	fprintf_method_declaration form m;
	fprintf form ")@]"
    | Typed_method_header(me,m) -> 
	fprintf form "@[<hv 1>Typed_method_header@,(%a,@ %a)@]"
	  Java_context.fprintf_method_entry me
	  fprintf_method_specification m
    | Constant_declaration(v) ->
	fprintf form "@[<hv 1>Constant_declaration@,(";
	fprintf_variable_declaration form v;
	fprintf form ")@]"
    | Typed_constant_declaration(v,init) ->
	fprintf form "@[<hv 1>Typed_constant_declaration@,(%a,@ %a)@]"
	  Java_context.fprintf_constant_entry v  
	  (*i Java_context.fprintf_instance_variable_entry v i*)
	  (fprintf_option fprintf_variable_initializer) init

  
let fprintf_commented_interface_member_declaration form (doc,decl) =
  fprintf form 
    "@[<hv 2>{ doc_comment =@ %a;@ decl =@ %a;@ }@]" 
    fprintf_javadoc doc
    fprintf_interface_member_declaration decl
;;

let fprintf_interface_declaration form d =
  fprintf form "@[<hv 2>{ @[<hv 2>interface_modifiers =@ ";
  fprintf_modifiers form d.interface_modifiers;
  fprintf form " ;@]@ @[<hv 2>interface_name =@ %a" 
    fprintf_identifier d.interface_name;
  fprintf form " ;@]@ @[<hv 2>interface_extends =@ ";
  fprintf_list fprintf_qualified_ident form d.interface_extends;
  fprintf form " ;@]@ @[<hv 2>interface_fields =@ ";
  fprintf_list fprintf_commented_interface_member_declaration form d.interface_members; 
  fprintf form "@]}@]"
;;

(*s compilation units *)



let fprintf_type_declaration form decl = 
  match decl with
    | Class(cd) ->
	fprintf form "@[<hv 1>Class@,(";
	fprintf_class_declaration form cd;
	fprintf form ")@]"
    | Interface(id) ->
	fprintf form "@[<hv 1>Interface@,(";
	fprintf_interface_declaration form id;	
	fprintf form ")@]"
    | Typed_class(ce,cd) ->
	fprintf form "@[<hv 1>Typed_class@,(%a,@ %a)@]"
	  Java_context.fprintf_class_entry ce
	  fprintf_class_declaration cd
    | Typed_interface(ie,id) ->
	fprintf form "@[<hv 1>Typed_interface@,(%a,@ %a)@]"
	  Java_context.fprintf_interface_entry ie	
	  fprintf_interface_declaration id	
;;  

let fprintf_commented_type_declaration form (doc,decl) =
  fprintf form 
    "@[<hv 2>{ doc_comment =@ %a;@ decl =@ %a;@ }@]" 
    fprintf_javadoc doc
    fprintf_type_declaration decl
;;




    

let fprintf_import_statement form i =
  match i with
    | Import_package(qi) ->
	fprintf form "Import_package@,(";
	fprintf_qualified_ident form qi;
	fprintf form ")"
    | Import_class_or_interface(qi) ->
	fprintf form "Import_class_or_interface@,(";
	fprintf_qualified_ident form qi;
	fprintf form ")"
;;



  
let fprintf_compilation_unit form c =
  fprintf form "@[<hv 2>{ @[<hv 2>package =@ ";
  fprintf_qualified_ident form c.cu_package;
  fprintf form " ;@]@ @[<hv 2>imports =@ ";
  fprintf_list fprintf_import_statement form c.cu_imports; 
  fprintf form " ;@]@ @[<hv 2>type_decls =@ ";
  fprintf_list fprintf_commented_type_declaration form c.cu_type_decls;
  fprintf form "@]}@]"
;; 

let printf_compilation_unit = fprintf_compilation_unit std_formatter
	
