
(** Krakatoa project - Model of objects using adresses in the heap *)

Require Import preambule.
Require Import classes.
Require Import ZArith.
Require Import map.

(** * Destructive model of objects *)
Set Implicit Arguments.
Unset Strict Implicit.
Module ObjectAdr (C: classes.ClassStructure).

Import C.
Module CGen := ClassGeneric C.
Import CGen.

Parameter adrObject : Set.

Parameter eq_adrObject_dec : forall a1 a2:adrObject, {a1 = a2} + {a1 <> a2}.

Inductive kind : Set :=
  | Obj : classId -> kind
  | Arr : Z -> javaType -> kind.

Definition eq_kind_dec : forall k1 k2:kind, {k1 = k2} + {k1 <> k2}.
decide equality.
apply eq_classId_dec.
apply eq_javaType_dec.
apply Z_eq_dec.
Defined.

Inductive value : Set :=
  | Primitive_bool : bool -> value
  | Primitive_int : Z -> value
  | Null : value
  | Ref : kind -> adrObject -> value.

Definition has_kind (v:value) (k:kind) : Prop :=
  match v with
  | Ref k1 _ => k = k1
  | _ => False
  end.


Definition default_val (t:javaType) : value :=
  match t with
  | IntT => Primitive_int 0
  | BoolT => Primitive_bool true
  | _ => Null
  end.

Definition neqv (v1 v2:value) := v1 <> v2.
Notation "x != y" := (neqv x y) (at level 70, no associativity).
Hint Unfold neqv.

Definition eq_value_dec : forall v1 v2:value, {v1 = v2} + {v1 != v2}.
unfold neqv in |- *; decide equality.
apply eq_bool_dec.
apply Z_eq_dec.
apply eq_adrObject_dec.
apply eq_kind_dec.
Defined.

(*
Definition eq_value : value -> value -> bool := 
  [x,y:value]
  if (eq_value_dec x y)
  then [_]true
  else [_]false.

Definition ne_value : value -> value -> bool := 
  [x,y:value] (negb(eq_value x y)).
*)

Definition eq_value :
  forall x y:value, {b : bool | if b then x = y else x <> y}.
Proof.
intros.
case (eq_value_dec x y); intros.
exists true; auto.
exists false; auto.
Qed.

Definition ne_value :
  forall x y:value, {b : bool | if b then x <> y else x = y}.
Proof.
intros.
case (eq_value_dec x y); intros.
exists false; auto.
exists true; auto.
Qed.

(** ** Access to Coq objects from Java values *)

Definition int_val (v:value) : Z :=
  match v with
  | Primitive_int n => n
  | _ => 0%Z (* assert false *)
  end.

Definition bool_val (v:value) : bool :=
  match v with
  | Primitive_bool b => b
  | _ => false (* assert false *)
  end.

Lemma int_val_simpl : forall n:Z, int_val (Primitive_int n) = n.
Proof.
trivial.
Qed.

Lemma bool_val_simpl : forall b:bool, bool_val (Primitive_bool b) = b.
Proof.
trivial.
Qed.

Coercion Primitive_int : Z >-> value.
Coercion Primitive_bool : bool >-> value.

Lemma int_val_eq : forall x y:value, x = y -> int_val x = int_val y.
simple destruct 1; trivial.
Qed.

Lemma bool_val_eq : forall x y:value, x = y -> bool_val x = bool_val y.
simple destruct 1; trivial.
Qed.

(*
Coercion int_val : value >-> Z.
Coercion bool_val : value >-> bool.
*)

(** ** Typing *)

Definition instanceof (v:value) (t:javaType) : Prop :=
  match v, t with
  | Primitive_int _, IntT => True
  | Primitive_bool _, BoolT => True
  | Null, ClassT _ => True
  | Null, ArrayT _ => True
  | Ref (Obj c1) _, ClassT c2 => extends c1 c2
  | Ref (Arr n t1) _, ArrayT t2 => (0 <= n)%Z /\ subType t1 t2
  | Ref (Arr n t1) _, ClassT c => (0 <= n)%Z /\ c = ObjectClass
  | _, _ => False
  end.

Lemma has_type_bool : forall b:bool, instanceof (Primitive_bool b) BoolType.
red in |- *; simpl in |- *; auto.
Qed.

Lemma has_type_int : forall n:Z, instanceof (Primitive_int n) IntType.
red in |- *; simpl in |- *; auto.
Qed.

Lemma has_type_null_obj : forall c:classId, instanceof Null (ClassType c).
red in |- *; simpl in |- *; auto.
Qed.

Lemma has_type_null_arr : forall t:javaType, instanceof Null (ArrayType t).
red in |- *; simpl in |- *; auto.
Qed.

Lemma has_type_obj :
 forall (c1 c2:classId) (adr:adrObject),
   extends c1 c2 -> instanceof (Ref (Obj c1) adr) (ClassType c2).
red in |- *; simpl in |- *; auto.
Qed.

Lemma has_type_arr :
 forall (l:Z) (adr:adrObject) (t1 t2:javaType),
   subType t1 t2 ->
   (0 <= l)%Z -> instanceof (Ref (Arr l t1) adr) (ArrayType t2).
red in |- *; simpl in |- *; auto.
Qed.


Hint Resolve has_type_int has_type_bool has_type_null_obj has_type_null_arr
  has_type_obj has_type_arr. 

Lemma prim_int_val_simpl :
 forall v:value, instanceof v IntType -> Primitive_int (int_val v) = v.
simple destruct v; simpl in |- *; intuition.
generalize H; case k; simpl in |- *; intuition.
Qed.

Lemma prim_bool_val_simpl :
 forall v:value, instanceof v BoolType -> Primitive_bool (bool_val v) = v.
unfold instanceof in |- *; simple destruct v; simpl in |- *; intuition.
generalize H; case k; simpl in |- *; intuition.
Qed.


Lemma has_type_subType_trans :
 forall (this:value) (t1:javaType),
   instanceof this t1 ->
   forall t2:javaType, subType t1 t2 -> instanceof this t2.
simple destruct this; simple destruct t1; simple destruct t2; simpl in |- *;
 intuition.
generalize H; case k; simpl in |- *; intuition.
apply extends_trans with c; trivial.
apply extends_Object; auto.
case H3; trivial.
generalize H; case k; simpl in |- *; intuition.
generalize H; case k; simpl in |- *; intuition.
apply subType_trans with j; trivial.
Qed.

Lemma has_type_default : forall t:javaType, instanceof (default_val t) t.
simple induction t; simpl in |- *; simpl in |- *; auto.
Qed.
Hint Resolve has_type_default.

Definition instanceof_dec :
  forall (v:value) (t:javaType), {instanceof v t} + {~ instanceof v t}.
simple destruct v; simple destruct t; try case k; simpl in |- *; intuition.
apply (extends_dec c c0).
case (eq_classId_dec c ObjectClass); intuition.
case (Z_le_dec 0 z); intuition.
case (subType_dec j j0); intuition.
case (Z_le_dec 0 z); intuition.
Defined.

(** Reference indexation *)

Inductive cell : Set :=
  | Field : fieldId -> cell
  | Index : Z -> cell.

Coercion Field : fieldId >-> cell.
Coercion Index : Z >-> cell.

Lemma eq_cell_dec : forall l1 l2:cell, {l1 = l2} + {l1 <> l2}.
decide equality.
apply eq_fieldId_dec.
apply Z_eq_dec.
Defined.

Lemma index_discriminate : forall i j:Z, i <> j -> Index i <> Index j.
intros; red in |- *; intros.
injection H0; auto.
Qed.

Lemma index_eq : forall i j:Z, i = j -> Index i = Index j.
simple destruct 1; trivial.
Qed.

Hint Resolve index_discriminate index_eq.

Definition is_kind_cell (k:kind) (c:cell) : Prop :=
  match k, c with
  | Obj c, Field f => extends c (defined_in f)
  | Arr n _, Index k => (0 <= k < n)%Z
  | _, _ => False
  end.

Definition is_kind_cell_dec :
  forall (k:kind) (c:cell), {is_kind_cell k c} + {~ is_kind_cell k c}.
intros; case k; case c; simpl in |- *; auto.
intros; apply extends_dec.
intros; apply Z_inf_dec; auto.
Defined.

Definition NullType := ClassType ObjectClass.

Definition type_of_cell (v:value) (c:cell) : javaType :=
  match c with
  | Field f => fieldType f
  | _ => match v with
         | Ref (Arr _ t) _ => t
         | _ => NullType
         end
  end.

Lemma type_of_cell_obj :
 forall (v:value) (f:fieldId), type_of_cell v f = fieldType f.
trivial.
Qed.


Lemma type_of_cell_array :
 forall (v:value) (n:Z) (t:javaType),
   v <> Null -> instanceof v (ArrayType t) -> subType (type_of_cell v n) t.
simple destruct v; simpl in |- *; try contradiction.
simpl in |- *; intuition.
simple destruct k; unfold instanceof in |- *; simpl in |- *; intuition.
Qed.

Lemma type_of_cell_array_int :
 forall (v:value) (n:Z),
   v <> Null ->
   instanceof v (ArrayType IntType) -> type_of_cell v n = IntType.
intros; apply subType_int; apply type_of_cell_array; trivial.
Qed.

Lemma type_of_cell_array_bool :
 forall (v:value) (n:Z),
   v <> Null ->
   instanceof v (ArrayType BoolType) -> type_of_cell v n = BoolType.
intros; apply subType_bool; apply type_of_cell_array; trivial.
Qed.

Definition is_valid_cell (v:value) (c:cell) : Prop :=
  match v with
  | Ref k _ => is_kind_cell k c
  | _ => False
  end.

Lemma is_valid_cell_not_null :
 forall (v:value) (c:cell), is_valid_cell v c -> v <> Null.
simple destruct v; intuition; try discriminate.
Qed.

Definition is_valid_cell_dec :
  forall (v:value) (c:cell), {is_valid_cell v c} + {~ is_valid_cell v c}.
simple destruct v; simpl in |- *; intros; auto.
apply is_kind_cell_dec.
Defined.

(** The reference values are represented as functions from cells to values *)
Definition cellMap := t cell value.

Definition cellMap_empty : cellMap := cte cell Null.

Definition cellMap_add : cellMap -> cell -> value -> cellMap :=
  add eq_cell_dec.

Definition cellMap_find : cellMap -> cell -> value :=
  find (I:=cell) (V:=value).

Lemma cellMap_find_add :
 forall (m:cellMap) (f:cell) (v:value),
   cellMap_find (cellMap_add m f v) f = v.
Proof find_add eq_cell_dec.

Lemma cellMap_find_add_eq :
 forall (m:cellMap) (f1 f2:cell) (v:value),
   f1 = f2 -> cellMap_find (cellMap_add m f1 v) f2 = v.
Proof find_add_eq eq_cell_dec.

Lemma cellMap_find_add_other :
 forall (m:cellMap) (f1 f2:cell) (v:value),
   f1 <> f2 -> cellMap_find (cellMap_add m f1 v) f2 = cellMap_find m f2.
Proof find_add_other eq_cell_dec.

(** The store is indexed by [kind] and [adrObject] *)

Record ref_record : Set := mk_rec {sort : kind; cells : adrObject}.

Definition val_ref (r:ref_record) : value := Ref (sort r) (cells r).

Definition eq_ref_record_dec :
  forall v1 v2:ref_record, {v1 = v2} + {v1 <> v2}.
decide equality.
apply eq_adrObject_dec.
apply eq_kind_dec.
Defined.

Lemma Ref_mk_rec_eq :
 forall (k1 k2:kind) (a1 a2:adrObject),
   Ref k1 a1 = Ref k2 a2 -> mk_rec k1 a1 = mk_rec k2 a2.
intros; injection H; intros; apply f_equal2 with (f := mk_rec); trivial.
Qed.

Lemma mk_rec_Ref_eq :
 forall (k1 k2:kind) (a1 a2:adrObject),
   mk_rec k1 a1 = mk_rec k2 a2 -> Ref k1 a1 = Ref k2 a2.
intros; injection H; intros; apply f_equal2 with (f := Ref); trivial.
Qed.

Hint Immediate Ref_mk_rec_eq mk_rec_Ref_eq.

Lemma Ref_mk_rec_neq :
 forall (k1 k2:kind) (a1 a2:adrObject),
   Ref k1 a1 <> Ref k2 a2 -> mk_rec k1 a1 <> mk_rec k2 a2.
red in |- *; intros; auto.
Qed.

Lemma mk_rec_Ref_neq :
 forall (k1 k2:kind) (a1 a2:adrObject),
   mk_rec k1 a1 <> mk_rec k2 a2 -> Ref k1 a1 <> Ref k2 a2.
red in |- *; intros; auto.
Qed.

Hint Immediate Ref_mk_rec_neq mk_rec_Ref_neq.

(** The store is defined as a partiel map from kind * adrObject to 
    map from cells to values *)

Definition rawstore := t ref_record (option cellMap).

(** Adding and finding references in the store *)
Definition rawstore_find_ref (s:rawstore) (o:ref_record) : cellMap :=
  if find s o then fun fs => fs else fun f => Null (* UndefinedRef *).

Definition rawstore_add_ref (s:rawstore) (o:ref_record) 
  (f:cellMap) : rawstore := add eq_ref_record_dec s o (Some f).

Lemma rawstore_find_add_eq :
 forall (s:rawstore) (r1 r2:ref_record) (c:cellMap),
   r1 = r2 -> rawstore_find_ref (rawstore_add_ref s r1 c) r2 = c.
intros; unfold rawstore_find_ref, rawstore_add_ref in |- *.
rewrite (find_add_eq eq_ref_record_dec s); trivial.
Qed.

Lemma rawstore_find_add_other :
 forall (s:rawstore) (r1 r2:ref_record) (c:cellMap),
   r1 <> r2 ->
   rawstore_find_ref (rawstore_add_ref s r1 c) r2 = rawstore_find_ref s r2.
intros; unfold rawstore_find_ref, rawstore_add_ref in |- *.
rewrite (find_add_other eq_ref_record_dec s); trivial.
Qed.

(** ** Access a cell of a value in the store *)

Definition rawaccess (s:rawstore) (this:value) (c:cell) : value :=
  match this with
  | Ref k o =>
      if is_kind_cell_dec k c
      then fun _ => cellMap_find (rawstore_find_ref s (mk_rec k o)) c
      else fun _ => Null (* NoSuchFieldException *)
  | Null => Null (* raise NullPointerException *)
  | _ => Null (* assert false *)
  end.

Lemma rawaccess_not_is_valid_cell :
 forall (s:rawstore) (v:value) (c:cell),
   ~ is_valid_cell v c -> rawaccess s v c = Null.
simple destruct v; simpl in |- *; intuition.
case (is_kind_cell_dec k c); intuition.
Qed.

(** Alived values *)
Definition rawalive (s:rawstore) (v:value) : Prop :=
  match v with
  | Ref k o => isdef (s (mk_rec k o))
  | _ => True
  end.

Lemma rawalive_dec :
 forall (s:rawstore) (v:value), {rawalive s v} + {~ rawalive s v}.
simple destruct v; simpl in |- *; auto.
intros k a; case (s (mk_rec k a)); auto.
Defined.


(** Property for an update to be correct *)
Definition rawis_valid_update (s:rawstore) (this:value) 
  (c:cell) (v:value) :=
  rawalive s v /\ is_valid_cell this c /\ instanceof v (type_of_cell this c).

Hint Unfold rawis_valid_update.

Lemma not_is_valid_cell_compat :
 forall (s:rawstore) (this:value) (c:cell) (v:value),
   ~ is_valid_cell this c -> ~ rawis_valid_update s this c v.
unfold rawis_valid_update in |- *; intuition.
Qed.

Lemma not_instanceof_compat :
 forall (s:rawstore) (this:value) (c:cell) (v:value),
   ~ instanceof v (type_of_cell this c) -> ~ rawis_valid_update s this c v.
unfold rawis_valid_update in |- *; intuition.
Qed.

Lemma rawis_valid_update_dec :
 forall (s:rawstore) (this:value) (c:cell) (v:value),
   {rawis_valid_update s this c v} + {~ rawis_valid_update s this c v}.
intros; unfold rawis_valid_update in |- *.
case (rawalive_dec s v); intuition.
case (instanceof_dec v (type_of_cell this c)); intuition.
case (is_valid_cell_dec this c); intuition.
Defined.

Definition rawupdate (s:rawstore) (this:value) (c:cell) 
  (v:value) : rawstore :=
  match this with
  | Ref k o =>
      if rawis_valid_update_dec s this c v
      then
       fun _ =>
         rawstore_add_ref s (mk_rec k o)
           (cellMap_add (rawstore_find_ref s (mk_rec k o)) c v)
      else fun _ => s (* ?? *)
  | Null => s (* raise NullPointerException *)
  | _ => s (* assert false *)
  end.

Lemma rawupdate_not_rawcompat_cell :
 forall (s:rawstore) (this:value) (c:cell) (v:value),
   ~ rawis_valid_update s this c v -> rawupdate s this c v = s.
simple destruct this; simpl in |- *; intuition.
elim (rawis_valid_update_dec s (Ref k a) c v) using dec_notA_elim; trivial.
Qed.

(** properties of [rawaccess] and [rawstore_add_ref] *)

Lemma rawaccess_add_ref_eq :
 forall (s:rawstore) (r1:ref_record) (v2:value) (c1:cellMap) 
   (c2:cell),
   val_ref r1 = v2 ->
   is_valid_cell v2 c2 -> rawaccess (rawstore_add_ref s r1 c1) v2 c2 = c1 c2.
simple destruct 1; case r1; simpl in |- *; try contradiction; intros; auto.
elim is_kind_cell_dec using dec_A_elim; simpl in |- *; intros; trivial.
rewrite rawstore_find_add_eq; trivial.
Qed.

Lemma rawaccess_add_other_ref :
 forall (s:rawstore) (r1:ref_record) (v2:value) (c1:cellMap) 
   (c2:cell),
   val_ref r1 <> v2 ->
   rawaccess (rawstore_add_ref s r1 c1) v2 c2 = rawaccess s v2 c2.
intros s r1 v2 c1 c2; case (is_valid_cell_dec v2 c2).
case r1; case v2; simpl in |- *; try contradiction; intros; auto.
elim is_kind_cell_dec using dec_A_elim; simpl in |- *; intros; trivial.
rewrite rawstore_find_add_other; trivial.
red in |- *; intro meq.
injection meq; intros; apply H.
unfold val_ref in |- *; simpl in |- *; apply f_equal2 with (f := Ref); auto.
intros; rewrite rawaccess_not_is_valid_cell; trivial.
intros; rewrite rawaccess_not_is_valid_cell; trivial.
Qed.

Lemma rawaccess_update :
 forall (s:rawstore) (this:value) (c:cell) (v:value),
   rawis_valid_update s this c v ->
   rawaccess (rawupdate s this c v) this c = v.
Proof.
unfold rawupdate, rawis_valid_update in |- *; simple destruct this;
 simpl in |- *; intuition.
elim (is_kind_cell_dec k c) using dec_A_elim; intros; trivial.
elim (rawis_valid_update_dec s (Ref k a) c v) using dec_A_elim; intros; auto.
unfold rawstore_add_ref, rawstore_find_ref in |- *.
rewrite (find_add eq_ref_record_dec s (mk_rec k a)).
rewrite cellMap_find_add; trivial.
Qed.

Lemma rawaccess_update_other :
 forall (s:rawstore) (this:value) (c1 c2:cell) (v:value),
   c1 <> c2 ->
   rawaccess (rawupdate s this c1 v) this c2 = rawaccess s this c2.
Proof.
intros s this c1 c2; case (is_valid_cell_dec this c2).
case this; simpl in |- *; intros; try contradiction.
elim (is_kind_cell_dec k c2) using dec_A_elim; intros; trivial.
case (rawis_valid_update_dec s (Ref k a) c1 v); intros; trivial.
unfold rawstore_add_ref, rawstore_find_ref in |- *; simpl in |- *.
rewrite (find_add eq_ref_record_dec s (mk_rec k a)).
apply cellMap_find_add_other; assumption.
intros; rewrite rawaccess_not_is_valid_cell; trivial.
rewrite rawaccess_not_is_valid_cell; trivial.
Qed.

Lemma rawaccess_update_other_val :
 forall (s:rawstore) (v1 v2:value) (c1 c2:cell) (v:value),
   v1 <> v2 -> rawaccess (rawupdate s v2 c2 v) v1 c1 = rawaccess s v1 c1.
Proof.
intros s v1 v2 c1; case (is_valid_cell_dec v1 c1).
case v1; simpl in |- *; intros; try contradiction.
elim (is_kind_cell_dec k c1) using dec_A_elim; intros; trivial.
generalize H; case v2; simpl in |- *; intros; try contradiction; trivial.
case (rawis_valid_update_dec s (Ref k0 a1) c2 v); intros; trivial.
unfold rawstore_add_ref, rawstore_find_ref in |- *; simpl in |- *.
rewrite (find_add_other eq_ref_record_dec s); trivial.
red in |- *; intro meq.
injection meq; intros; apply H0.
apply f_equal2 with (f := Ref); auto.
intros; rewrite rawaccess_not_is_valid_cell; trivial.
rewrite rawaccess_not_is_valid_cell; trivial.
Qed.


(** Well-formed stores *)
Lemma rawalive_new_stable :
 forall (s:rawstore) (v:value) (r:ref_record) (cm:cellMap),
   rawalive s v -> rawalive (rawstore_add_ref s r cm) v.
simple destruct v; simple destruct r; simpl in |- *; intros; auto.
unfold rawstore_add_ref in |- *; auto.
unfold add in |- *.
elim (eq_ref_record_dec (mk_rec sort0 cells0) (mk_rec k a)); trivial.
red in |- *; auto.
Qed.

Lemma rawalive_update_stable :
 forall (s:rawstore) (v v1 v2:value) (c:cell),
   rawalive s v -> rawalive (rawupdate s v1 c v2) v.
simple destruct v; simpl in |- *; intuition.
unfold rawupdate in |- *; case v1; intros; auto.
case (rawis_valid_update_dec s (Ref k0 a0) c v2); intros; auto.
unfold rawstore_add_ref, add in |- *; auto.
case (eq_ref_record_dec (mk_rec k0 a0) (mk_rec k a)); auto.
Qed.

Definition rawwell_formed (s:rawstore) (v:value) (t:javaType) :=
  rawalive s v /\ instanceof v t.

Lemma rawwell_formed_stable :
 forall s s':rawstore,
   (forall v:value, rawalive s v -> rawalive s' v) ->
   forall (v:value) (t:javaType),
     rawwell_formed s v t -> rawwell_formed s' v t.
unfold rawwell_formed in |- *; intuition.
Qed.


Definition store_well_formed (s:rawstore) : Prop :=
  forall (v:value) (c:cell),
    is_valid_cell v c ->
    rawwell_formed s (rawaccess s v c) (type_of_cell v c).

Record store : Set := mk_store
  {raw : rawstore; rawwell : store_well_formed raw}.

Definition access (s:store) (this:value) (c:cell) : value :=
  rawaccess (raw s) this c.

Lemma access_not_is_valid_cell :
 forall (s:store) (this:value) (c:cell),
   ~ is_valid_cell this c -> access s this c = Null.
unfold access in |- *; intros; apply rawaccess_not_is_valid_cell; trivial.
Qed.

Lemma store_add_ref :
 forall (s:store) (r:ref_record) (cm:cellMap),
   (forall c:cell,
      is_valid_cell (val_ref r) c ->
      rawwell_formed (raw s) (cm c) (type_of_cell (val_ref r) c)) -> store.
intros; exists (rawstore_add_ref (raw s) r cm); red in |- *; intros.
case (eq_value_dec v (val_ref r)); intros.

rewrite rawaccess_add_ref_eq; subst v; auto.
apply rawwell_formed_stable with (raw s); auto.
intros; apply rawalive_new_stable; auto.

rewrite rawaccess_add_other_ref; auto.
case s; simpl in |- *; unfold store_well_formed in |- *; intros.
apply rawwell_formed_stable with raw0; auto.
intros; apply rawalive_new_stable; auto.
Defined.
Implicit Arguments store_add_ref [].

(** The creation of new addresses is axiomatised *)
Parameter new_addr : store -> kind -> adrObject.

Definition new_rec (s:store) (k:kind) : ref_record := mk_rec k (new_addr s k).

Definition new_ref (s:store) (k:kind) : value := Ref k (new_addr s k).

Definition new_array (s:store) (n:Z) (t:javaType) : value :=
  if Z_le_dec 0 n then fun _ => new_ref s (Arr n t) else fun _ => Null.

Definition new_obj (s:store) (c:classId) : value := new_ref s (Obj c).

Axiom is_new_addr : forall (s:store) (k:kind), undef (raw s (new_rec s k)).

Lemma is_valid_cell_new :
 forall (s:store) (k:kind) (c:cell),
   is_kind_cell k c -> is_valid_cell (new_ref s k) c.
auto.
Qed.

Lemma is_valid_cell_new_array :
 forall (s:store) (n:Z) (t:javaType) (c:cell),
   is_kind_cell (Arr n t) c ->
   (0 <= n)%Z -> is_valid_cell (new_array s n t) c.
unfold new_array in |- *; intros.
elim (Z_le_dec 0 n) using dec_A_elim; auto.
Qed.

Lemma is_valid_cell_new_obj :
 forall (s:store) (d:classId) (c:cell),
   is_kind_cell (Obj d) c -> is_valid_cell (new_obj s d) c.
auto.
Qed.

Lemma is_valid_cell_new_obj_eq :
 forall (s:store) (v:value) (c:classId) (f:cell),
   v = new_obj s c -> is_kind_cell (Obj c) f -> is_valid_cell v f.
intuition; subst; auto.
Qed.

Lemma new_is_diff :
 forall (s:store) (k:kind) (a:ref_record),
   isdef (raw s a) -> a <> new_rec s k.
red in |- *; intros.
subst a.
absurd (isdef (raw s (new_rec s k))); trivial.
generalize (is_new_addr s k); case (raw s (new_rec s k));
 unfold undef in |- *; simpl in |- *; intros; auto.
discriminate H0.
Qed.

Definition is_new_ref (s:store) (v:value) :=
  match v with
  | Ref k o => undef (raw s (mk_rec k o))
  | _ => False
  end.

Lemma new_ref_is_new : forall (s:store) (k:kind), is_new_ref s (new_ref s k).
intros; simpl in |- *.
unfold new_rec in |- *; exact (is_new_addr s k).
Qed.

Lemma new_obj_well_typed :
 forall (s:store) (c:classId), instanceof (new_obj s c) (ClassType c).
red in |- *; simpl in |- *; auto.
Qed.

Lemma new_obj_well_typed_eq :
 forall (s:store) (c:classId) (v:value),
   v = new_obj s c -> instanceof v (ClassType c).
intros; subst v; red in |- *; simpl in |- *; auto.
Qed.

Lemma new_array_well_typed :
 forall (s:store) (z:Z) (t:javaType),
   instanceof (new_array s z t) (ArrayType t).
intros s z t; unfold new_array in |- *; case (Z_le_dec 0 z); intro.
red in |- *; simpl in |- *; intuition.
apply subType_refl.
trivial.
Qed.


(** The cellMap to be allocated when creating an object in kind [k] *)
Definition new_ref_map (k:kind) : cellMap :=
  fun c =>
    match k, c with
    | Obj id, Field f =>
        if extends_dec id (defined_in f)
        then fun _ => default_val (fieldType f)
        else fun _ => Null
    | Arr m t, Index n =>
        if Z_inf_dec n m then fun _ => default_val t else fun _ => Null
    | _, _ => Null
    end.

Lemma default_val_alive :
 forall (s:rawstore) (t:javaType), rawalive s (default_val t).
simple destruct t; simpl in |- *; auto.
Qed.

Hint Resolve default_val_alive.

Lemma new_ref_map_well_formed :
 forall (s:store) (k:kind) (a:adrObject) (c:cell),
   is_valid_cell (Ref k a) c ->
   rawwell_formed (raw s) (new_ref_map k c) (type_of_cell (Ref k a) c).
simple destruct k. 
simple destruct c0; simpl in |- *; intros; auto.
elim extends_dec using dec_A_elim; trivial; split; auto.
split; simpl in |- *; auto.
simple destruct c; simpl in |- *; intros; try contradiction; auto.
elim Z_inf_dec using dec_A_elim; trivial; split; simpl in |- *; auto.
Qed.
Implicit Arguments new_ref_map_well_formed [].


Definition store_new_ref (s:store) (v:value) : store :=
  match v with
  | Ref k a =>
      store_add_ref s (mk_rec k a) (new_ref_map k)
        (new_ref_map_well_formed s k a)
  | _ => s
  end.

Lemma update : store -> value -> cell -> value -> store.
intros s this c v; exists (rawupdate (raw s) this c v).
case s; simpl in |- *; intros.
case (rawis_valid_update_dec raw0 this c v); intro.
case (eq_value_dec this Null); intros.
subst this; simpl in |- *; trivial.
red in |- *; intros.
case (eq_value_dec this v0); intro.
case (eq_cell_dec c c0); intro.
subst this c; rewrite rawaccess_update; red in r; intuition.
red in |- *; intuition.
apply rawalive_update_stable; auto.

subst this; rewrite rawaccess_update_other; auto.
case (rawwell0 v0 c0); red in |- *; intuition.
apply rawalive_update_stable; auto.

rewrite rawaccess_update_other_val; auto.
case (rawwell0 v0 c0); red in |- *; intuition.
apply rawalive_update_stable; auto.

rewrite rawupdate_not_rawcompat_cell; trivial.
Defined.


Lemma rawaccess_update_other_cell :
 forall (s:rawstore) (v1 v2:value) (c1 c2:cell) (v:value),
   c1 <> c2 -> rawaccess (rawupdate s v2 c2 v) v1 c1 = rawaccess s v1 c1.
intros; case (eq_value_dec v1 v2); intro.
subst v1; apply rawaccess_update_other; auto.
apply rawaccess_update_other_val; auto.
Qed.

(** Definition of compat/alive on stores *)

Definition alive (s:store) := rawalive (raw s).
Definition fresh (s:store) (v:value) := ~ alive s v.
Definition is_valid_update (s:store) := rawis_valid_update (raw s).
Definition well_formed (s:store) := rawwell_formed (raw s).

Lemma well_formed_instanceof :
 forall (s:store) (this:value) (t:javaType),
   well_formed s this t -> instanceof this t.
simple destruct 1; auto.
Qed.

Lemma well_formed_alive :
 forall (s:store) (this:value) (t:javaType),
   well_formed s this t -> alive s this.
simple destruct 1; auto.
Qed.

Lemma well_formed_new_diff :
 forall (s:store) (v1 v2:value) (t:javaType),
   well_formed s v1 t -> fresh s v2 -> v1 != v2.
Proof.
simple destruct 1; unfold fresh, alive in |- *; red in |- *; intuition.
subst; auto.
Qed.

Lemma well_formed_intro :
 forall (s:store) (this:value) (t:javaType),
   alive s this -> instanceof this t -> well_formed s this t.
split; auto.
Qed.

(** Properties of [update] and [access] *)

Lemma access_update :
 forall (s:store) (this:value) (c:cell) (v:value),
   is_valid_update s this c v -> access (update s this c v) this c = v.
Proof.
unfold access, update in |- *; simpl in |- *.
intros; apply rawaccess_update; trivial.
Qed.

(** An equivalent version useful for automatic proofs *)
Lemma access_update_eq :
 forall (s:store) (v1 v2:value) (c1 c2:cell) (v:value),
   v1 = v2 ->
   c1 = c2 ->
   is_valid_update s v1 c1 v -> access (update s v1 c1 v) v2 c2 = v.
simple destruct 1; simple destruct 1; intros.
apply access_update; trivial.
Defined.

Lemma access_update_other :
 forall (s:store) (this:value) (c1 c2:cell) (v:value),
   c1 <> c2 -> access (update s this c1 v) this c2 = access s this c2.
Proof.
unfold access, update in |- *; simpl in |- *.
intros; apply rawaccess_update_other; trivial.
Qed.

Lemma access_update_other_val :
 forall (s:store) (v1 v2:value) (c1 c2:cell) (v:value),
   v1 <> v2 -> access (update s v2 c2 v) v1 c1 = access s v1 c1.
Proof.
unfold access, update in |- *; simpl in |- *.
intros; apply rawaccess_update_other_val; trivial.
Qed.


Lemma access_update_other_cell :
 forall (s:store) (v1 v2:value) (c1 c2:cell) (v:value),
   c1 <> c2 -> access (update s v2 c2 v) v1 c1 = access s v1 c1.
Proof.
intros; case (eq_value_dec v1 v2); intro.
subst v1; apply access_update_other; auto.
apply access_update_other_val; auto.
Qed.

Lemma access_update_neq :
 forall (s:store) (v1 v2:value) (c1 c2:cell) (v:value),
   c1 <> c2 \/ v1 <> v2 -> access (update s v2 c2 v) v1 c1 = access s v1 c1.
Proof.
simple destruct 1; intros.
apply access_update_other_cell; trivial.
apply access_update_other_val; trivial.
Qed.

(* Specialized version for more efficient automatic rewriting *)
Lemma access_index_update_field :
 forall (s:store) (v1 v2:value) (z:Z) (f:fieldId) (v:value),
   access (update s v2 f v) v1 z = access s v1 z.
intros; rewrite access_update_neq; auto.
Qed.

Lemma access_field_update_index :
 forall (s:store) (v1 v2:value) (z:Z) (f:fieldId) (v:value),
   access (update s v2 z v) v1 f = access s v1 f.
intros; rewrite access_update_neq; auto.
Qed.


(** New objects are not alive in the store *)

(*
Lemma new_obj_is_new_old :
 (s:store)(c:classId)(v:value)(alive s v)->~v=(new_obj s c).
Unfold new_obj new_ref; Destruct v; Simpl; Intros; Try Discriminate.
LetTac lem := (new_is_diff 2!(Obj c) H).
Red; Intro Heq; Injection Heq; Intros.
Apply lem.
Subst a k; Auto.
Save.
*)

Lemma new_obj_is_new :
 forall (s:store) (c:classId) (v1 v2:value),
   alive s v1 -> v2 = new_obj s c -> v1 <> v2.
unfold new_obj, new_ref in |- *; simple destruct v1; simpl in |- *; intros;
 subst v2; try (red in |- *; discriminate).
set (lem := new_is_diff (k:=(Obj c)) H) in *.
red in |- *; intro Heq; injection Heq; intros.
apply lem.
subst a k; auto.
Qed.

(*
Lemma new_obj_is_new_sym_old :
 (s:store)(c:classId)(v:value)(alive s v)->~(new_obj s c)=v.
Intros; Apply sym_not_eq; Apply new_obj_is_new; Trivial.
Save.
*)

Lemma new_array_is_new :
 forall (s:store) (t:javaType) (n:Z) (v:value),
   (0 <= n)%Z -> alive s v -> v <> new_array s n t.
unfold new_array, new_ref in |- *; simple destruct v; simpl in |- *; intros;
 elim (Z_le_dec 0 n) using dec_A_elim; intros; trivial; 
 try discriminate.
set (lem := new_is_diff (k:=(Arr n t)) H0) in *.
red in |- *; intro Heq; injection Heq; intros.
apply lem.
subst a k; auto.
Qed.

Lemma new_array_is_new_sym :
 forall (s:store) (t:javaType) (n:Z) (v:value),
   (0 <= n)%Z -> alive s v -> new_array s n t <> v.
intros; apply sym_not_eq; apply new_array_is_new; trivial.
Qed.

Lemma new_obj_not_alive :
 forall (s:store) (c:classId) (v:value), v = new_obj s c -> fresh s v.
red in |- *; intros s c v H H1; apply (new_obj_is_new (c:=c) (v2:=v) H1);
 trivial.
Qed.

Lemma new_array_not_alive :
 forall (s:store) (t:javaType) (n:Z), (0 <= n)%Z -> fresh s (new_array s n t).
intros s t n npos H; apply (new_array_is_new (t:=t) (n:=n) npos H); trivial.
Qed.

Hint Resolve new_obj_not_alive new_array_not_alive.

Lemma alive_update :
 forall (s:store) (v v1 v2:value) (c:cell),
   alive s v -> alive (update s v1 c v2) v.
unfold alive in |- *; simpl in |- *; intros.
apply rawalive_update_stable; trivial.
Qed.

Lemma alive_update_eq :
 forall (s s':store) (v v1 v2:value) (c:cell),
   alive s v -> s' = update s v1 c v2 -> alive s' v.
intros s s' v v1 v2 c H H1; rewrite H1; apply alive_update; trivial.
Qed.

Lemma alive_new_ref :
 forall (s:store) (k:kind),
   alive (store_new_ref s (new_ref s k)) (new_ref s k).
red in |- *; unfold store_new_ref in |- *; simpl in |- *; intros.
unfold rawstore_add_ref in |- *; auto.
unfold add in |- *.
elim (eq_ref_record_dec (mk_rec k (new_addr s k)) (mk_rec k (new_addr s k)))
 using dec_A_elim; auto.
Qed.

Lemma alive_new_obj :
 forall (s:store) (c:classId),
   alive (store_new_ref s (new_obj s c)) (new_obj s c).
Proof fun s c => alive_new_ref s (Obj c).

Lemma alive_new_array :
 forall (s:store) (z:Z) (t:javaType),
   alive (store_new_ref s (new_array s z t)) (new_array s z t).
intros; unfold new_array in |- *; case (Z_le_dec 0 z); auto.
intro; apply alive_new_ref.
simpl in |- *; auto.
Qed.

Lemma alive_new_stable :
 forall (s:store) (v v1:value), alive s v -> alive (store_new_ref s v1) v.
simple destruct v; simple destruct v1; simpl in |- *; intros; auto.
unfold rawstore_add_ref in |- *; auto.
unfold add in |- *.
elim (eq_ref_record_dec (mk_rec k0 a0) (mk_rec k a)); trivial.
red in |- *; auto.
Qed.

Lemma alive_new_stable_eq :
 forall (s s':store) (v v1:value),
   alive s v -> s' = store_new_ref s v1 -> alive s' v.
intros s s' v v1 H H1; rewrite H1; apply alive_new_stable; trivial.
Qed.

(** * Properties of well-formed *)

Lemma well_formed_update_eq :
 forall (s s':store) (v v1 v2:value) (c:cell) (t:javaType),
   well_formed s v t -> s' = update s v1 c v2 -> well_formed s' v t.
intros; apply well_formed_intro.
apply alive_update_eq with s v1 v2 c; trivial.
apply well_formed_alive with t; trivial.
apply well_formed_instanceof with s; trivial.
Qed.

Lemma well_formed_update :
 forall (s:store) (v v1 v2:value) (c:cell) (t:javaType),
   well_formed s v t -> well_formed (update s v1 c v2) v t.
intros; apply well_formed_update_eq with s v1 v2 c; trivial.
Qed.

Lemma well_formed_new_obj_old :
 forall (s:store) (c:classId),
   well_formed (store_new_ref s (new_obj s c)) (new_obj s c) (ClassType c).
intros; apply well_formed_intro; auto.
apply alive_new_obj.
apply new_obj_well_typed.
Qed.

Lemma well_formed_new_obj :
 forall (s1 s2:store) (new:value) (c:classId),
   new = new_obj s1 c ->
   s2 = store_new_ref s1 new -> well_formed s2 new (ClassType c).
intros; apply well_formed_intro; auto.
subst s2 new; apply alive_new_obj.
subst new; apply new_obj_well_typed.
Qed.

Lemma well_formed_new_array :
 forall (s:store) (z:Z) (t:javaType),
   well_formed (store_new_ref s (new_array s z t)) (
     new_array s z t) (ArrayType t).
intros; apply well_formed_intro.
apply alive_new_array.
apply new_array_well_typed.
Qed.


Lemma well_formed_new_stable :
 forall (s:store) (v v1:value) (t:javaType),
   well_formed s v t -> well_formed (store_new_ref s v1) v t.
intros; apply well_formed_intro.
apply alive_new_stable.
apply well_formed_alive with t; trivial.
apply well_formed_instanceof with s; trivial.
Qed.

Lemma well_formed_new_stable_eq :
 forall (s s':store) (v v1:value) (t:javaType),
   well_formed s v t -> s' = store_new_ref s v1 -> well_formed s' v t.
intros s s' v v1 t H H1; rewrite H1; apply well_formed_new_stable; trivial.
Qed.

Lemma well_formed_int :
 forall (s:store) (v:value), instanceof v IntType -> well_formed s v IntType.
intros; apply well_formed_intro; trivial.
generalize H; case v; simpl in |- *; auto.
simple destruct k; intuition.
Qed.

Lemma well_formed_bool :
 forall (s:store) (v:value),
   instanceof v BoolType -> well_formed s v BoolType.
intros; apply well_formed_intro; trivial.
generalize H; case v; simpl in |- *; auto.
simple destruct k; intuition.
Qed.

Definition non_prim (t:javaType) :=
  match t with
  | IntT => False
  | BoolT => False
  | _ => True
  end.

Lemma well_formed_null :
 forall (s:store) (t:javaType), non_prim t -> well_formed s Null t.
intros; apply well_formed_intro; trivial.
generalize H; case t; simpl in |- *; intuition.
Qed.

(** * Properties of [access] *)

Lemma access_new_ref_other :
 forall (s:store) (v1 v2:value) (c:cell),
   v1 <> v2 -> access (store_new_ref s v1) v2 c = access s v2 c.
intros s v1 v2 c; case (is_valid_cell_dec v2 c).
case v1; case v2; try contradiction; simpl in |- *; intros; auto.
elim is_kind_cell_dec using dec_A_elim; intros; trivial.
intros; rewrite rawstore_find_add_other; auto.
intros; rewrite access_not_is_valid_cell; trivial.
rewrite access_not_is_valid_cell; trivial.
Qed.

Lemma access_new_ref_eq :
 forall (s:store) (v1 v2:value) (c:cell),
   v1 = v2 ->
   is_valid_cell v2 c ->
   access (store_new_ref s v1) v2 c = default_val (type_of_cell v2 c).
simple destruct 1; case v1; simpl in |- *; try contradiction; auto.
simple destruct k; case c; auto.
intros; elim is_kind_cell_dec using dec_A_elim; simpl in |- *; intros;
 trivial.
rewrite rawstore_find_add_eq; auto.
unfold cellMap_find, find in |- *.
unfold new_ref_map in |- *; auto.
elim (extends_dec c0 (defined_in f)) using dec_A_elim; simpl in |- *; intros;
 trivial.
intros; elim is_kind_cell_dec using dec_A_elim; simpl in |- *; intros;
 trivial.
unfold cellMap_find, find in |- *.
rewrite rawstore_find_add_eq; auto.
unfold new_ref_map in |- *; simpl in |- *; intros; auto; try contradiction.
intros; elim is_kind_cell_dec using dec_A_elim; simpl in |- *; intros;
 trivial.
unfold cellMap_find, find in |- *.
rewrite rawstore_find_add_eq; auto.
unfold new_ref_map in |- *; simpl in |- *; intros; auto; try contradiction.
elim (Z_inf_dec z z0) using dec_A_elim; trivial.
Qed.

Definition arraylength (this:value) : Z :=
  match this with
  | Ref (Arr n _) _ => n
  | Null => 0%Z (* raise NullPointerException *)
  | _ => 0%Z (* never happens on well-typed programs *)
  end.


Lemma arraylength_new_array :
 forall (s:store) (n:Z) (t:javaType) (v:value),
   v = new_array s n t -> (n >= 0)%Z -> arraylength v = n.
unfold arraylength, new_array, new_ref in |- *; intros s n.
case (Z_le_dec 0 n); intuition.
subst v; auto.
Qed.

Lemma arraylength_always_pos :
 forall (v:value) (t:javaType),
   instanceof v (ArrayType t) -> (0 <= arraylength v)%Z.
simple destruct v; simpl in |- *; intuition.
generalize H; case k; intuition.
Qed.

Definition obj_class (this:value) : option classId :=
  match this with
  | Ref (Obj c) _ => Some c
  | Null => None (* raise NullPointerException *)
  | _ => None (* never happens on well-typed programs *)
  end.


Lemma has_type_obj_inv :
 forall (s:store) (this:value) (f:fieldId) (t:javaType),
   is_valid_cell this f -> fieldType f = t -> instanceof (access s this f) t.
simple destruct s; simple destruct this; auto; try contradiction.
simple destruct k; intros; auto; try contradiction.
rewrite <- H0; case (rawwell0 (Ref (Obj c) a) f H); intuition.
Qed.

Lemma has_type_array_inv :
 forall (s:store) (this:value) (t:javaType) (n:Z),
   instanceof this (ArrayType t) ->
   is_valid_cell this n -> instanceof (access s this n) t.
Proof.
simple destruct s; simple destruct this; auto; try contradiction.
simple destruct k; intros; auto; try contradiction.
red in H0.
simpl in H0.
case (rawwell0 (Ref (Arr z j) a) n H0).
unfold type_of_cell in |- *; intuition.
apply has_type_subType_trans with j; auto.
case H; trivial.
Qed.


Lemma compat_access_field :
 forall (s:store) (v1 v2:value) (f1:fieldId),
   is_valid_cell v1 f1 ->
   alive s v2 -> instanceof v2 (fieldType f1) -> is_valid_update s v1 f1 v2.
red in |- *; simpl in |- *; intuition.
Qed.

(*
Lemma compat_access_cell : (s:store)(v1,v2:value;z1,z2:Z)
     (is_valid_cell v1 z1)->(is_valid_cell v2 z2)
    ->(instanceof v2 (static_type v1))
    ->(is_valid_update v1 z1 (access s v2 z2)).
Destruct v1; Simpl; Red; Try Contradiction.
Destruct k; Simpl; Intuition.
Apply has_type_array_inv; Trivial.
Save.

Lemma instance_of_array_int_static : 
   (v:value)(~v=Null)->(instanceof v (ArrayType IntType))
                     -> (static_type v)=(ArrayType IntType).
Destruct v; Unfold instanceof; Simpl; Try Contradiction.
Intuition.
Destruct k; Simpl; Try Contradiction.
Destruct j; Intuition.
Save.
*)

Lemma compat_access_cell_int :
 forall (s:store) (v1 v2:value) (z1:Z),
   is_valid_cell v1 z1 ->
   instanceof v1 (ArrayType IntType) ->
   instanceof v2 IntType -> is_valid_update s v1 z1 v2.
simple destruct v1; simpl in |- *; red in |- *; try contradiction.
simple destruct k; simpl in |- *; intuition.
assert (alive s v2).
generalize H1; case v2; simpl in |- *; auto.
simple destruct k0; contradiction.
generalize H4; case j; red in |- *; simpl in |- *; intuition.
Qed.

Lemma compat_access_cell_bool :
 forall (s:store) (v1 v2:value) (z1:Z),
   is_valid_cell v1 z1 ->
   instanceof v1 (ArrayType BoolType) ->
   instanceof v2 BoolType -> is_valid_update s v1 z1 v2.
simple destruct v1; simpl in |- *; red in |- *; try contradiction.
simple destruct k; simpl in |- *; intuition.
assert (alive s v2).
generalize H1; case v2; simpl in |- *; auto.
simple destruct k0; contradiction.
generalize H4; case j; red in |- *; simpl in |- *; intuition.
Qed.

Lemma is_valid_cell_has_type_obj :
 forall (this:value) (f:fieldId),
   this <> Null ->
   instanceof this (ClassType (defined_in f)) -> is_valid_cell this f.
Proof.
simple destruct this; intros; auto.
case H; trivial.
generalize H0; case k; auto.
unfold instanceof in |- *; simpl in |- *; intros.
absurd (defined_in f = ObjectClass); auto.
apply no_field_in_Object.
intuition.
Qed.

Lemma is_valid_cell_has_type_array :
 forall (this:value) (n:Z),
   (0 <= n)%Z -> (n < arraylength this)%Z -> is_valid_cell this n.
Proof.
unfold is_valid_cell, arraylength in |- *; simple destruct this; intros;
 try omega.
unfold is_kind_cell in |- *; induction  k as [c| z j]; omega.
Qed.

(** Values of incompatible types are different *)

Lemma subType_comparable :
 forall t1 t2 t3:javaType,
   subType t1 t2 -> subType t1 t3 -> subType t2 t3 \/ subType t3 t2.
simple induction t1; simple destruct t2; simple destruct t3; simpl in |- *;
 intuition.
apply extends_comparable with c; trivial.
rewrite H0; rewrite H1; auto.
Qed.

Lemma instanceof_comparable :
 forall (t1 t2:javaType) (v:value),
   v <> Null ->
   instanceof v t1 -> instanceof v t2 -> subType t1 t2 \/ subType t2 t1.
simple destruct t1; simple destruct t2; simple destruct v; simpl in |- *;
 try simple destruct k; intuition.
apply extends_comparable with c1; trivial.
subst c0 c; auto.
apply subType_comparable with j1; auto.
Qed.

Definition not_comparable (t1 t2:javaType) :=
  ~ subType t1 t2 /\ ~ subType t2 t1.

Lemma not_comparable_sym :
 forall t1 t2:javaType, not_comparable t1 t2 -> not_comparable t2 t1.
unfold not_comparable in |- *; intuition.
Qed.

Lemma not_comparable_int_bool : not_comparable IntType BoolType.
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_int_class :
 forall c:classId, not_comparable IntType (ClassType c).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_bool_class :
 forall c:classId, not_comparable BoolType (ClassType c).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_int_array :
 forall t:javaType, not_comparable IntType (ArrayType t).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_bool_array :
 forall t:javaType, not_comparable BoolType (ArrayType t).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_class_array :
 forall (c:classId) (t:javaType),
   c <> ObjectClass -> not_comparable (ClassType c) (ArrayType t).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_array_class :
 forall (c:classId) (t:javaType),
   c <> ObjectClass -> not_comparable (ArrayType t) (ClassType c).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_class :
 forall c1 c2:classId,
   ~ extends c1 c2 ->
   ~ extends c2 c1 -> not_comparable (ClassType c1) (ClassType c2).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_array :
 forall t1 t2:javaType,
   not_comparable t1 t2 -> not_comparable (ArrayType t1) (ArrayType t2).
unfold not_comparable in |- *; simpl in |- *; intuition.
Qed.

Lemma not_comparable_type_diff :
 forall (t1 t2:javaType) (v1 v2:value),
   v1 <> Null \/ v2 <> Null ->
   instanceof v1 t1 -> instanceof v2 t2 -> not_comparable t1 t2 -> v1 <> v2.
unfold not_comparable in |- *; red in |- *; intros; auto.
case (instanceof_comparable (t1:=t1) (t2:=t2) (v:=v1)); intuition.
apply H2; auto.
rewrite <- H3; auto.
rewrite H3; auto.
rewrite H3; auto.
Qed.

(** Modifiable *)

Definition mod_loc := value -> cell -> Prop.

Definition nothing_loc : mod_loc := fun v c => True.
Definition everything_loc : mod_loc := fun v c => False.

Lemma nothing_intro : forall (v:value) (c:cell), nothing_loc v c.
red in |- *; auto.
Qed.

Lemma not_everything : forall (v:value) (c:cell), ~ everything_loc v c.
unfold everything_loc in |- *; intuition.
Qed.

Definition inter_loc (l1 l2:mod_loc) v c := l1 v c /\ l2 v c.

Lemma inter_intro :
 forall (l1 l2:mod_loc) (v:value) (c:cell),
   l1 v c -> l2 v c -> inter_loc l1 l2 v c.
red in |- *; auto.
Qed.

Lemma not_inter :
 forall (l1 l2:mod_loc) (v:value) (c:cell),
   ~ l1 v c \/ ~ l2 v c -> ~ inter_loc l1 l2 v c.
unfold inter_loc in |- *; intuition.
Qed.

Definition access_loc (v1:value) (c1:cell) : mod_loc :=
  fun v c => v1 <> v \/ c1 <> c.

Lemma access_intro :
 forall (v1 v:value) (c1 c:cell), v1 <> v \/ c1 <> c -> access_loc v1 c1 v c.
red in |- *; trivial.
Qed.

Lemma not_access_refl : forall (v:value) (c:cell), ~ access_loc v c v c.
unfold access_loc in |- *; intuition.
Qed.

Lemma not_access_eq :
 forall (v1 v:value) (c1 c:cell), v1 = v -> c1 = c -> ~ access_loc v c v c.
unfold access_loc in |- *; intuition.
Qed.

Definition allcells_loc (v1:value) : mod_loc := fun v c => v1 <> v.

Lemma allcells_intro :
 forall (v1 v:value) (c:cell), v1 <> v -> allcells_loc v1 v c.
red in |- *; auto.
Qed.

Lemma not_allcells_refl : forall (v:value) (c:cell), ~ allcells_loc v v c.
unfold allcells_loc in |- *; intuition.
Qed.

Lemma not_allcells_eq :
 forall (v1 v:value) (c:cell), v1 = v -> ~ allcells_loc v v c.
unfold allcells_loc in |- *; intuition.
Qed.


(* Set of locations [v,c] which cannot be accessed from value [v1] *)
CoInductive unreachable_ind (s:store) (v:value) (c:cell) : value -> Prop :=
    UnReachableIntro :
      forall v1:value,
        v1 <> v /\
        (forall c1:cell,
           is_valid_cell v1 c1 -> unreachable_ind s v c (access s v1 c1)) ->
        unreachable_ind s v c v1.

Definition reachable_loc (s:store) (v1:value) : mod_loc :=
  fun v c => unreachable_ind s v c v1.

Lemma reachable_intro :
 forall (s:store) (v1 v:value) (c:cell),
   v1 <> v ->
   (forall c1:cell,
      is_valid_cell v1 c1 -> reachable_loc s (access s v1 c1) v c) ->
   reachable_loc s v1 v c.
intros; red in |- *; apply UnReachableIntro; auto.
Qed.

Lemma not_reachable_refl :
 forall (s:store) (v:value) (c:cell), ~ reachable_loc s v v c.
unfold reachable_loc in |- *; intuition.
inversion H; intuition.
Qed.

Lemma not_reachable_eq :
 forall (s:store) (v1 v:value) (c:cell), v1 = v -> ~ reachable_loc s v1 v c.
unfold reachable_loc in |- *; intuition.
inversion H0; intuition.
Qed.

Lemma not_reachable_access :
 forall (s:store) (v v1:value) (c c1:cell),
   is_valid_cell v1 c1 ->
   ~ reachable_loc s (access s v1 c1) v c -> ~ reachable_loc s v1 v c.
unfold reachable_loc in |- *; intuition.
inversion H1; intuition.
Qed.


Definition array_sub_loc (v1:value) (z1 z2:Z) : mod_loc :=
  fun v c =>
    v1 <> v \/
    match c with
    | Index i => (i < z1)%Z \/ (z2 < i)%Z
    | _ => True
    end.

Lemma array_sub_intro_val :
 forall (v1:value) (z1 z2:Z) (v:value) (c:cell),
   v1 <> v -> array_sub_loc v1 z1 z2 v c.
unfold array_sub_loc in |- *; intuition.
Qed.

Lemma array_sub_intro_index :
 forall (v v1:value) (z1 z2 i:Z),
   (i < z1)%Z \/ (z2 < i)%Z -> array_sub_loc v1 z1 z2 v i.
unfold array_sub_loc in |- *; intuition.
Qed.

Lemma not_array_sub_refl :
 forall (v:value) (z1 z2 i:Z),
   (z1 <= i <= z2)%Z -> ~ array_sub_loc v z1 z2 v i.
unfold array_sub_loc in |- *; intuition.
Qed.

Lemma not_array_sub_eq :
 forall (v v1:value) (z1 z2 i:Z),
   v = v1 -> (z1 <= i <= z2)%Z -> ~ array_sub_loc v1 z1 z2 v i.
unfold array_sub_loc in |- *; intuition.
Qed.

Definition modifiable (s s':store) (unchanged:mod_loc) : Prop :=
  forall v:value,
    alive s v ->
    alive s' v /\
    (forall c:cell, unchanged v c -> access s v c = access s' v c).


Lemma not_alive_access :
 forall (s:store) (v:value) (c:cell), fresh s v -> access s v c = Null.
unfold fresh in |- *; simple destruct v; simpl in |- *; intuition.
case (is_kind_cell_dec k c); intuition.
unfold rawstore_find_ref in |- *.
unfold find in |- *.
generalize H; case (raw s (mk_rec k a)); simpl in |- *; intuition.
Qed.

Lemma alive_access :
 forall (s:store) (this:value) (c:cell),
   is_valid_cell this c -> alive s (access s this c).
intros [raws wf] this c H.
case (wf this c); intuition.
Qed.

Lemma mod_access_rev :
 forall (s s':store) (unchanged:mod_loc) (v:value) 
   (c:cell),
   modifiable s s' unchanged ->
   unchanged v c -> alive s v -> access s v c = access s' v c.
intros s s' mod_ v c H H1.
generalize (H v); intuition.
Qed.

Lemma mod_access :
 forall (s s':store) (unch:mod_loc) (v:value) (c:cell),
   modifiable s s' unch ->
   unch v c -> alive s v -> access s' v c = access s v c.
intros s s' unch v c; symmetry  in |- *;
 apply (mod_access_rev (s:=s) (s':=s') (unchanged:=unch) (v:=v) (c:=c));
 trivial.
Qed.

Lemma mod_alive :
 forall (s s':store) (unchanged:mod_loc) (v:value),
   modifiable s s' unchanged -> alive s v -> alive s' v.
intros s s' mod_ v H.
generalize (H v); intuition.
Qed.

Lemma mod_well_formed :
 forall (s s':store) (unchanged:mod_loc) (v:value) 
   (t:javaType),
   modifiable s s' unchanged -> well_formed s v t -> well_formed s' v t.
intros s s' mod_ v t H H1.
generalize (well_formed_instanceof (s:=s) (this:=v) (t:=t));
 generalize (well_formed_alive (s:=s) (this:=v) (t:=t)).
intros; apply well_formed_intro; generalize (H v); intuition.
Qed.

Lemma mod_implies :
 forall (s s':store) (unch unch':mod_loc),
   (forall (v:value) (c:cell), unch' v c -> unch v c) ->
   modifiable s s' unch -> modifiable s s' unch'.
red in |- *; intros; intuition.
apply mod_alive with (1 := H0); trivial.
apply mod_access_rev with (1 := H0); auto.
Qed.

Lemma mod_inter_left :
 forall (s s':store) (unch unch':mod_loc),
   modifiable s s' unch -> modifiable s s' (inter_loc unch unch').
intros; apply mod_implies with (2 := H).
unfold inter_loc in |- *; intuition.
Qed.

Lemma mod_inter_right :
 forall (s s':store) (unch unch':mod_loc),
   modifiable s s' unch' -> modifiable s s' (inter_loc unch unch').
intros; apply mod_implies with (2 := H).
unfold inter_loc in |- *; intuition.
Qed.

Lemma mod_refl : forall (s:store) (unch:mod_loc), modifiable s s unch.
red in |- *; intuition.
Qed.

Lemma mod_update_eq :
 forall (s:store) (unch:mod_loc) (v e:value) (c:cell),
   ~ unch v c -> modifiable s (update s v c e) unch.
red in |- *; intuition.
apply alive_update; trivial.
symmetry  in |- *; apply access_update_neq; auto.
case (eq_cell_dec c0 c); intuition.
case (eq_value_dec v0 v); intuition.
absurd (unch v c); trivial; case e0; case e1; trivial.
Qed.

Lemma mod_everything :
 forall s s':store,
   (forall v:value, alive s v -> alive s' v) ->
   modifiable s s' everything_loc.
red in |- *; unfold everything_loc in |- *; intuition.
Qed.

Lemma mod_trans :
 forall (s s' s'':store) (unch:mod_loc),
   modifiable s s' unch -> modifiable s' s'' unch -> modifiable s s'' unch.
unfold modifiable in |- *; intros.
generalize (H v); generalize (H0 v); intuition.
transitivity (access s' v c); auto.
Qed.

Lemma mod_update :
 forall (s s':store) (unch:mod_loc) (v e:value) (c:cell),
   modifiable s s' unch -> ~ unch v c -> modifiable s (update s' v c e) unch.
intros; apply mod_trans with s'; trivial.
apply mod_update_eq; trivial.
Qed.

Lemma mod_update_everything :
 forall (s s':store) (v e:value) (c:cell),
   modifiable s s' everything_loc ->
   modifiable s (update s' v c e) everything_loc.
intros; apply mod_update; auto.
Qed.

Lemma mod_new_stable :
 forall (s s':store) (unch:mod_loc) (v:value),
   modifiable s s' unch ->
   fresh s' v -> modifiable s (store_new_ref s' v) unch.
unfold modifiable in |- *; intros.
case (H v0); intuition.
apply alive_new_stable; auto.
rewrite access_new_ref_other; auto.
red in |- *; intro Heq; apply H0; rewrite Heq; auto.
Qed.

Lemma mod_new_everything :
 forall (s s':store) (v:value),
   modifiable s s' everything_loc ->
   modifiable s (store_new_ref s' v) everything_loc.
unfold modifiable, everything_loc in |- *; intros.
case (H v0); intuition.
apply alive_new_stable; auto.
Qed.

Lemma mod_implies_everything :
 forall (s s':store) (unch:mod_loc),
   modifiable s s' unch -> modifiable s s' everything_loc.
unfold modifiable, everything_loc in |- *; intuition.
case (H v); intuition.
Qed.
Lemma mod_trans_gen :
 forall (s1 s2 s3:store) (unch1 unch2 unch3:mod_loc),
   modifiable s1 s2 unch1 ->
   modifiable s2 s3 unch2 ->
   (forall (v:value) (c:cell),
      alive s1 v -> unch3 v c -> unch1 v c /\ unch2 v c) ->
   modifiable s1 s3 unch3.
unfold modifiable in |- *; intros.
case (H v); trivial; intros.
case (H0 v); trivial; intuition.
case (H1 v c); trivial; intros.
transitivity (access s2 v c); intuition.
Qed.

Lemma mod_trans_allcells :
 forall (s1 s2 s3:store) (unch1:mod_loc) (v:value),
   fresh s1 v ->
   modifiable s1 s2 unch1 ->
   modifiable s2 s3 (allcells_loc v) -> modifiable s1 s3 unch1.
intros; apply mod_trans_gen with s2 unch1 (allcells_loc v); auto.
intuition.
repeat red in |- *; intros.
subst v; auto.
Qed.

Lemma mod_after_new :
 forall (s1 s2 s3:store) (new:value) (c:classId),
   new = new_obj s1 c ->
   s2 = store_new_ref s1 new ->
   modifiable s2 s3 (allcells_loc new) -> modifiable s1 s3 nothing_loc.
Proof.
unfold modifiable in |- *; intros.
assert (alive s2 v).
subst s2; apply alive_new_stable; assumption.
split.
case (H1 v); auto.
intros.
assert (v <> new).
subst new.
apply new_obj_is_new with s1 c; trivial.
transitivity (access s2 v c0).
subst s2.
rewrite access_new_ref_other; auto.
case (H1 v); intros; auto.
apply H7; red in |- *; auto.
Qed.

End ObjectAdr.

