
(** Krakatoa project - Functional model of objects *)



Require Import preambule.
Require Import classes.
(* Require classes_gen. *)
Require Import ZArith.
Require Import map.

Set Implicit Arguments.
Unset Strict Implicit.

(** * Functional model of objects *)

Module ObjectVal (C: classes.ClassStructure).

Import C.
Module CGen := ClassGeneric C.
Import CGen.

(** ** Co-inductive definition of values 
    an object is represented by its class 
    and a (partial) function from fields name to value *)

(*
Could be defined from a generic concrete value 
CoInductive value : Set := 
   Val : (cvalue classId (fieldId-> value)) -> value.
*)

CoInductive value : Set :=
  | Primitive_bool : bool -> value
  | Primitive_int : Z -> value
  | Null : value
  | Ref : object_record -> value
  | Array : array_struct -> value
with object_record : Set :=
    Build_object :
      forall (the_class:classId) (the_fields:fieldId -> value), object_record
with array_struct : Set :=
    Build_array : forall (length:Z) (cells:Z -> value), array_struct.

(** ** Useful abreviations for object manipulation *)

Definition fieldMap := t fieldId value.

Definition the_class (o:object_record) : classId :=
  match o with
  | Build_object c f => c
  end.

Definition the_fields (o:object_record) : fieldMap :=
  match o with
  | Build_object c f => f
  end.

Definition FieldMap_empty : fieldMap := cte fieldId Null.

Definition FieldMap_add : fieldMap -> fieldId -> value -> fieldMap :=
  add eq_fieldId_dec.

Definition FieldMap_find : fieldMap -> fieldId -> value :=
  find (I:=fieldId) (V:=value).

Lemma FieldMap_find_add :
 forall (m:fieldMap) (f:fieldId) (v:value),
   FieldMap_find (FieldMap_add m f v) f = v.
Proof find_add eq_fieldId_dec.

Lemma FieldMap_find_add_other :
 forall (m:fieldMap) (f1 f2:fieldId) (v:value),
   f1 <> f2 -> FieldMap_find (FieldMap_add m f1 v) f2 = FieldMap_find m f2.
Proof find_add_other eq_fieldId_dec.


(** Array indexation *)

Definition array_struct_length (t:array_struct) : Z :=
  match t with
  | Build_array l c => l
  end.

Definition array_struct_cell (t:array_struct) : Z -> value :=
  match t with
  | Build_array l c => c
  end.

Definition arrayMap := t Z value.

Definition ArrayMap_empty : arrayMap := cte Z Null.

Definition ArrayMap_add : arrayMap -> Z -> value -> arrayMap := add Z_eq_dec.

Definition ArrayMap_find : arrayMap -> Z -> value := find (I:=Z) (V:=value).

Lemma ArrayMap_find_add :
 forall (m:arrayMap) (n:Z) (v:value),
   ArrayMap_find (ArrayMap_add m n v) n = v.
Proof find_add Z_eq_dec.

Lemma ArrayMap_find_add_other :
 forall (m:arrayMap) (n1 n2:Z) (v:value),
   n1 <> n2 -> ArrayMap_find (ArrayMap_add m n1 v) n2 = ArrayMap_find m n2.
Proof find_add_other Z_eq_dec.


(** ** Access to Coq objects from Java values *)
Definition int_val (v:value) : Z :=
  match v with
  | Primitive_int n => n
  | _ => 0%Z (* assert false *)
  end.

Definition int_val_meth := int_val.

Definition bool_val (v:value) : bool :=
  match v with
  | Primitive_bool b => b
  | _ => false (* assert false *)
  end.

Definition bool_val_meth := bool_val.

Lemma int_val_simpl : forall n:Z, int_val (Primitive_int n) = n.
Proof.
trivial.
Qed.

Lemma bool_val_simpl : forall b:bool, bool_val (Primitive_bool b) = b.
Proof.
trivial.
Qed.


(** ** [instanceof] relation *)
(**  [(instanceof v c)] is true if [v] is a value which inherits from 
     the class [c].
*)

Definition instanceof (v:value) (c:classId) : Prop :=
  match v with
  | Ref o => extends (the_class o) c
  | _ => False (* assert false *)
  end.

Lemma instanceof_inv :
 forall (v:value) (c:classId),
   instanceof v c ->
    exists2 o : object_record , v = Ref o & extends (the_class o) c.
Proof.
intro v.
case v; unfold instanceof in |- *; simpl in |- *; try contradiction.
intros.
exists o; trivial.
Qed.

(** [instanceof] is a decidable relation, the corresponding boolean function 
    is [instanceof_bool] *)

Definition instanceof_bool (v:value) (c:classId) : bool :=
  match v with
  | Ref o => extends_bool (the_class o) c
  | _ => false (* assert false *)
  end.


Lemma instanceof_bool_inv :
 forall (v:value) (c:classId),
   instanceof_bool v c = true ->
    exists2 o : object_record , v = Ref o &
   extends_bool (the_class o) c = true.
Proof.
intro v.
case v; unfold instanceof_bool in |- *; simpl in |- *; intros;
 try discriminate.
intros.
exists o; trivial.
Qed.

(** ** Java [cast] function *)

Definition cast (i:classId) (v:value) : value :=
  match v with
  | Null => Null (* and not NullPointerException ! see Java semantics *)
  | Ref o =>
      if extends_bool (the_class o) i
      then v
      else Null (* raise ClassCastException *)
  | _ => Null (* assert false *)
  end.
  
(** ** Access to fields *)

Definition access (this:value) (f:fieldId) : value :=
  match this with
  | Ref this_o =>
      if exists_in_bool f (the_class this_o)
      then FieldMap_find (the_fields this_o) f
      else Null (* NoSuchFieldException *)
  | Null => Null (* raise NullPointerException *)
  | _ => Null (* assert false *)
  end.


(** ** Update of fields *)

Definition update (this:value) (f:fieldId) (v:value) : value :=
  match this with
  | Ref this_o =>
      if exists_in_bool f (the_class this_o)
      then
       Ref
         (Build_object (the_class this_o)
            (FieldMap_add (the_fields this_o) f v))
      else Null (* NoSuchFieldException *)
  | Null => Null (* raise NullPointerException *)
  | _ => Null (* assert false *)
  end.

(** ** Useful lemmas on [access] and [update] *)

Lemma access_update :
 forall (this:value) (f:fieldId) (v:value),
   instanceof this (defined_in f) -> access (update this f v) f = v.
Proof.
intros.
elim (instanceof_inv H); intros.
subst this.
unfold access, update in |- *; simpl in |- *.
unfold exists_in_bool in |- *.
assert (H1bool := extends_Prop_to_bool (the_class x) (defined_in f) H1).
repeat (rewrite H1bool; simpl in |- *).
apply FieldMap_find_add.
Qed.

Lemma instanceof_update :
 forall (this:value) (f:fieldId) (c:classId) (v:value),
   extends c (defined_in f) ->
   instanceof this c -> instanceof (update this f v) c.
Proof.
intros this.
case this; simpl in |- *; try contradiction.
intros o f c v H1 H2.
unfold exists_in_bool in |- *;
 rewrite (extends_Prop_to_bool (the_class o) (defined_in f)).
unfold instanceof in |- *; assumption.
apply extends_trans with c; trivial.
Qed.

Lemma access_update_other :
 forall (this:value) (f1 f2:fieldId) (v:value),
   instanceof this (defined_in f1) ->
   f1 <> f2 -> access (update this f1 v) f2 = access this f2.
Proof.
intros.
unfold access, update in |- *; simpl in |- *.
elim (instanceof_inv H); intros.
subst this.
unfold exists_in_bool in |- *.
rewrite (extends_Prop_to_bool (the_class x) (defined_in f1) H2);
 simpl in |- *.
case (extends_bool (the_class x) (defined_in f2)); trivial.
apply FieldMap_find_add_other; assumption.
Qed.

(** ** specializes versions of access/update *)

Definition access_int (f:fieldId) (this:value) : Z := int_val (access this f).

Definition update_int (f:fieldId) (this:value) (n:Z) : value :=
  update this f (Primitive_int n).

Lemma access_update_int :
 forall (this:value) (f:fieldId) (n:Z),
   instanceof this (defined_in f) -> access_int f (update_int f this n) = n.
unfold access_int, update_int in |- *.
intros; rewrite access_update; auto.
Qed.

Lemma access_update_int_other :
 forall (this:value) (f1 f2:fieldId) (n:Z),
   instanceof this (defined_in f1) ->
   f1 <> f2 -> access (update_int f1 this n) f2 = access this f2.
unfold access_int, update_int in |- *.
intros; rewrite access_update_other; auto.
Qed.

Definition access_bool (f:fieldId) (this:value) : bool :=
  bool_val (access this f).

Definition update_bool (f:fieldId) (this:value) (n:bool) : value :=
  update this f (Primitive_bool n).

Lemma access_update_bool :
 forall (this:value) (f:fieldId) (n:bool),
   instanceof this (defined_in f) -> access_bool f (update_bool f this n) = n.
unfold access_bool, update_bool in |- *.
intros; rewrite access_update; auto.
Qed.

Lemma access_update_bool_other :
 forall (this:value) (f1 f2:fieldId) (n:bool),
   instanceof this (defined_in f1) ->
   f1 <> f2 -> access (update_bool f1 this n) f2 = access this f2.
unfold access_bool, update_bool in |- *.
intros; rewrite access_update_other; auto.
Qed.

(** Arrays *)

Definition isarray (v:value) : Prop :=
  match v with
  | Array t => True
  | _ => False (* assert false *)
  end.

Lemma isarray_inv :
 forall v:value, isarray v ->  exists t : array_struct , v = Array t.
Proof.
intro v.
case v; unfold isarray in |- *; simpl in |- *; try contradiction.
intros.
exists a; trivial.
Qed.

Definition array_length (this:value) : Z :=
  match this with
  | Array t => array_struct_length t
  | Null => 0%Z (* raise NullPointerException *)
  | _ => 0%Z (* never happens on well-typed programs *)
  end.


Definition array_length_meth := array_length.

Definition array_access (this:value) (n:Z) : value :=
  match this with
  | Array t =>
      if Z_ge_lt_dec n 0
      then
       fun _ =>
         if Z_gt_le_dec (array_struct_length t) n
         then fun _ => array_struct_cell t n
         else fun _ => Null (* ArrayOutOfBoundsException *)
      else fun _ => Null (* ArrayOutOfBoundsException *)
  | Null => Null (* raise NullPointerException *)
  | _ => Null (* never happens on well-typed programs *)
  end.

Definition array_access_meth := array_access.


(** Update of arrays *)

Definition array_fun_update (t:Z -> value) (n:Z) (v:value) 
  (k:Z) : value := if Z_eq_dec n k then fun _ => v else fun _ => t k.
 
Definition array_update (this:value) (n:Z) (v:value) : value :=
  match this with
  | Array (Build_array l t) =>
      if Z_ge_lt_dec n 0
      then
       fun _ =>
         if Z_gt_le_dec l n
         then fun _ => Array (Build_array l (array_fun_update t n v))
         else fun _ => Null (* ArrayOutOfBoundsException *)
      else fun _ => Null (* ArrayOutOfBoundsException *)
  | Null => Null (* raise NullPointerException *)
  | _ => Null (* assert false *)
  end.

Definition int_array_update (this:value) (n v:Z) : value :=
  array_update this n (Primitive_int v).

(** ** Useful lemmas on [array_access] and [array_update] *)

Lemma array_length_update :
 forall (this:value) (n:Z) (v:value),
   isarray this ->
   (0 <= n)%Z ->
   (n < array_length this)%Z ->
   array_length (array_update this n v) = array_length this.
Proof.
intros this n v H.
elim (isarray_inv H).
intros x H1.
subst this.
case x; simpl in |- *; intros.
case (Z_ge_lt_dec n 0); intros; try omega.
case (Z_gt_le_dec length n); intros; try omega.
trivial.
Qed.

Lemma isarray_update :
 forall (this:value) (n:Z) (v:value),
   isarray this ->
   (0 <= n)%Z -> (n < array_length this)%Z -> isarray (array_update this n v).
Proof.
intros this n v H.
elim (isarray_inv H).
intros x H1.
subst this.
case x; simpl in |- *; intros.
case (Z_ge_lt_dec n 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n); intros; try (elimtype False; omega).
trivial.
Qed.

Lemma array_access_update_same :
 forall (this:value) (n:Z) (v:value),
   isarray this ->
   (0 <= n)%Z ->
   (n < array_length this)%Z -> array_access (array_update this n v) n = v.
Proof.
intros this n v H.
elim (isarray_inv H).
intros x H1.
subst this.
case x; simpl in |- *; intros.
case (Z_ge_lt_dec n 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n); intros; try (elimtype False; omega).
simpl in |- *.
case (Z_ge_lt_dec n 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n); intros; try (elimtype False; omega).
unfold array_fun_update in |- *.
case (Z_eq_dec n n); intros; try (elimtype False; omega).
trivial.
Qed.

Lemma array_access_update_other :
 forall (this:value) (n1 n2:Z) (v:value),
   isarray this ->
   (0 <= n1)%Z ->
   (n1 < array_length this)%Z ->
   (0 <= n2)%Z ->
   (n2 < array_length this)%Z ->
   n1 <> n2 ->
   array_access (array_update this n1 v) n2 = array_access this n2.
Proof.
intros this n1 n2 v H.
elim (isarray_inv H).
intros x H1.
subst this.
case x; simpl in |- *; intros.
case (Z_ge_lt_dec n1 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n1); intros; try (elimtype False; omega).
case (Z_ge_lt_dec n2 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n2); intros; try (elimtype False; omega).
simpl in |- *.
case (Z_ge_lt_dec n2 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n2); intros; try (elimtype False; omega).
unfold array_fun_update in |- *.
case (Z_eq_dec n1 n2); intros; try (elimtype False; omega).
trivial.
Qed.

Lemma array_access_update :
 forall (this:value) (n1 n2:Z) (v:value),
   isarray this ->
   (0 <= n1)%Z ->
   (n1 < array_length this)%Z ->
   (0 <= n2)%Z ->
   (n2 < array_length this)%Z ->
   array_access (array_update this n1 v) n2 =
   (if Z_eq_dec n1 n2 then fun _ => v else fun _ => array_access this n2).
Proof.
intros this n1 n2 v H.
elim (isarray_inv H).
intros x H1.
subst this.
case x; simpl in |- *; intros.
case (Z_ge_lt_dec n1 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n1); intros; try (elimtype False; omega).
case (Z_ge_lt_dec n2 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n2); intros; try (elimtype False; omega).
simpl in |- *.
case (Z_ge_lt_dec n2 0); intros; try (elimtype False; omega).
case (Z_gt_le_dec length n2); intros; try (elimtype False; omega).
trivial.
Qed.

(** * Default values, new objects *)

Definition default (t:javaType) : value :=
  match t with
  | IntT => Primitive_int 0
  | BoolT => Primitive_bool true
  | _ => Null
  end.

Definition new (c:classId) : value :=
  Ref
    (Build_object c
       (fun f => if exists_in_bool f c then default (fieldType f) else Null)).

Lemma instanceof_new : forall c:classId, instanceof (new c) c.
simpl in |- *; auto.
Qed.


(** new arrays *)

Definition build_array (len:Z) (init:value) : value :=
  Array
    (Build_array len
       (fun i =>
          if Z_ge_lt_dec i 0
          then
           fun _ =>
             if Z_gt_le_dec len i then fun _ => init else fun _ => Null
          else fun _ => Null)).
    
    

(** ** Equality on values *)

CoInductive eq_value : value -> value -> Prop :=
  | eq_val_refl : forall v:value, eq_value v v
  | eq_val_ref :
      forall (c:classId) (o1 o2:fieldMap),
        (forall f:fieldId, eq_value (o1 f) (o2 f)) ->
        eq_value (Ref (Build_object c o1)) (Ref (Build_object c o2))
  | eq_val_array :
      forall (l:Z) (t1 t2:arrayMap),
        (forall n:Z, eq_value (t1 n) (t2 n)) ->
        eq_value (Array (Build_array l t1)) (Array (Build_array l t2)).

Definition ne_value (x y:value) : Prop := ~ eq_value x y.


Hint Constructors eq_value.
Lemma eq_value_trans :
 forall v1 v2 v3:value, eq_value v1 v2 -> eq_value v2 v3 -> eq_value v1 v3.
cofix.
simple destruct 1.
trivial.
intros.
inversion H1.
apply eq_val_ref; trivial.
apply eq_val_ref; intro.
apply eq_value_trans with (o2 f); trivial.
intros.
inversion H1.
apply eq_val_array; trivial.
apply eq_val_array; intro.
apply eq_value_trans with (t2 n); trivial.
Qed.

Lemma eq_value_sym : forall v1 v2:value, eq_value v1 v2 -> eq_value v2 v1.
cofix.
simple destruct 1.
exact eq_val_refl.
intros; apply eq_val_ref; intros.
apply eq_value_sym; trivial.
intros; apply eq_val_array; intros.
apply eq_value_sym; trivial.
Qed.


End ObjectVal.
