
(** Krakatoa project - Preambule *)

(** * Useful parameterisation of Coq *)

Implicit Arguments Some [A].
Implicit Arguments None [A].
Implicit Arguments well_founded [A].
Implicit Arguments Acc [A].
Implicit Arguments Fix [A R].
Hint Extern 1 (_ <> _) => discriminate.

Definition eq_bool_dec : forall b1 b2:bool, {b1 = b2} + {b1 <> b2}.
decide equality.
Defined.

(** * Additional definitions and lemmas *)

(** ** The [option] type *)

Set Implicit Arguments.
Unset Strict Implicit.

Section option_def.
Variable A : Set.

Definition undef (x:option A) : Prop := x = None.

Definition def (x:option A) (y:A) : Prop := x = Some y.

Definition isdef (x:option A) : Prop :=
  match x with
  | Some _ => True
  | None => False
  end.

Hint Unfold def undef.

Lemma undefNone : undef None.
auto.
Qed.

Lemma defSome : forall x:A, def (Some x) x.
auto.
Qed.

Inductive def_dec (x:option A) : Set :=
  | case_def : forall y:A, def x y -> def_dec x
  | case_undef : undef x -> def_dec x.

Hint Constructors def_dec.

Definition is_def : forall x:option A, def_dec x.
simple destruct x; eauto.
Defined.

Lemma def_fun : forall (x:option A) (y z:A), def x y -> def x z -> y = z.
unfold def in |- *; intros.
assert (Some y = Some z).
transitivity x; auto.
inversion H1; trivial.
Defined.

Lemma defSome_eq : forall x y:A, def (Some x) y -> x = y.
intros; apply def_fun with (Some x); auto.
Defined.

Lemma undef_not_def : forall (x:option A) (y:A), undef x -> ~ def x y.
red in |- *; simple destruct x; unfold def, undef in |- *; intros.
discriminate H.
discriminate H0.
Qed.

End option_def.
Hint Unfold def undef isdef.
Hint Immediate defSome_eq.

(* ** Decidable properties *)

Section Decidability.

Variable A : Prop.

Lemma dec_A_elim :
 A ->
 forall P:{A} + {~ A} -> Type,
   (forall a:A, P (left (~ A) a)) -> forall s:{A} + {~ A}, P s.
simple destruct s; intros; auto.
case n; trivial.
Qed.

Lemma dec_notA_elim :
 ~ A ->
 forall P:{A} + {~ A} -> Type,
   (forall a:~ A, P (right A a)) -> forall s:{A} + {~ A}, P s.
simple destruct s; intros; auto.
case H; trivial.
Qed.

Variable A_dec : {A} + {~ A}.

Definition dec_bool : bool := if A_dec then fun _ => true else fun _ => false.
Definition dec_prop : Prop := if A_dec then fun _ => True else fun _ => False.
Definition not_dec_prop : Prop :=
  if A_dec then fun _ => False else fun _ => True.

Lemma dec_A_proof : dec_prop -> A.
unfold dec_prop in |- *; case A_dec; intuition.
Qed.

Lemma dec_not_A_proof : not_dec_prop -> ~ A.
unfold not_dec_prop in |- *; case A_dec; intuition.
Qed.

Lemma dec_bool_true_intro : A -> dec_bool = true.
intros; unfold dec_bool in |- *; elim A_dec using dec_A_elim; intros; auto.
Qed.

Lemma dec_bool_true_elim : dec_bool = true -> A.
unfold dec_bool in |- *; case A_dec; intros; auto.
discriminate H; trivial.
Qed.

Lemma dec_bool_false_intro : ~ A -> dec_bool = false.
intros; unfold dec_bool in |- *; elim A_dec using dec_notA_elim; intros; auto.
Qed.

Lemma dec_bool_false_elim : dec_bool = false -> ~ A.
unfold dec_bool in |- *; case A_dec; intros; auto.
discriminate H; trivial.
Qed.

End Decidability.
Implicit Arguments dec_bool_false_elim [A].
Implicit Arguments dec_bool_true_elim [A].
Implicit Arguments dec_A_proof [A].
Implicit Arguments dec_not_A_proof [A].

Require Import ZArith.

Lemma Z_inf_dec : forall n m:Z, {(0 <= n < m)%Z} + {~ (0 <= n < m)%Z}.
intros; case (Z_le_dec 0 n); intros.
intros; case (Z_lt_dec n m); intros; auto.
right; omega.
right; omega.
Defined.

Lemma Ncompare_eq : forall n:N, (n ?= n)%N = Datatypes.Eq.
simple destruct n; simpl in |- *; auto.
intro; apply Pcompare_refl.
Qed.

Lemma entier_eq_dec : forall n m:N, {n = m} + {n <> m}.
intros n m; case (Dcompare_inf (n ?= m)%N); intro.
case s; intro.
left; apply Ncompare_Eq_eq; auto.
right; red in |- *; intro; subst.
assert (Datatypes.Eq = Datatypes.Lt); try discriminate.
rewrite <- (Ncompare_eq m); trivial.
right; red in |- *; intro; subst.
assert (Datatypes.Eq = Datatypes.Gt); try discriminate.
rewrite <- (Ncompare_eq m); trivial.
Qed.


Lemma Z_entier_pos :
 forall n:Z, (0 <= n)%Z -> BinInt.Z_of_N (BinInt.Zabs_N n) = n.
simple destruct n; simpl in |- *; intros; auto with *.
absurd (0 <= Zneg p)%Z; auto with zarith.
Qed.


(* (nat_sigma n a f) = sigma_{i=0}^{n} f(a+i) *)
Fixpoint nat_sigma (n:nat) (a:Z) (f:Z -> Z) {struct n} : Z :=
  match n with
  | O => f a
  | S p => (f a + nat_sigma p (a + 1) f)%Z
  end.

(* Eval Compute in (nat_sigma 5 `0` ([z:Z]z)). 

  doit faire 15
*)

(* (Zsigma a b f) = sigma_{i=a}^{b} f(i) *)

Definition Zsigma (a b:Z) (f:Z -> Z) : Z :=
  match (b - a)%Z with
  | Zneg _ => 0%Z
  | Z0 => f a
  | Zpos p => nat_sigma (nat_of_P p) a f
  end.

