

open Ast_types;;
open Ast;;

exception Desugaring_failure of string;;

let desugaring_failure msg =
  raise (Desugaring_failure msg)

let type_of_Exception = Type_name [Location.dummy, "Exception";
				   Location.dummy, "lang";
				   Location.dummy, "java"]


(* [expr_list_to_conj l] returns a single clause which is a conjunction 
   of all clauses in list [l] in the following fashion:
\[ c_1; c_2; \cdots ; c_n \longrightarrow c_1 \&\& c_2 \&\& \ldots \&\& c_n \].
   Each localisation component of the original clauses are kept;
   localisation components of the added conjunction expressions 
   is [dummy]. *)

let expr_list_to_conj l =
  match l with
    | [] -> expr_true
    | clause::r -> 
	mk_expr 
	  clause.expr_loc
	  ((List.fold_left 
	      (fun acc e -> mk_expr Location.dummy (Bin(acc,"&&",e))) 
	      clause 
	      r).expr_node)
;;


(* [rename exc_type old new_e expr] returns expression [expr] where free 
   occurrences of variable name [old] are replaced by variable name [new_e].
   Several cases are unused due to programming style hypothesis. *)


let rec rename exc_type old new_e expr =
  { expr_loc = expr.expr_loc;
    expr_node= 
      match expr.expr_node with
	| Lit l -> Lit l
	| Bin(e1,s,e2) -> Bin(rename exc_type old new_e e1,
                              s,
                              rename exc_type old new_e e2)
	| Un(s,e1) -> Un(s,rename exc_type old new_e e1)
	| PrePostIncrDecr(s,e1) -> PrePostIncrDecr(s,rename exc_type old new_e e1)
	| (Var(loc,id)) as v -> if id = old then Cast(exc_type , 
                                                      mk_expr Location.dummy 
							(Var(loc,new_e))) 
	  else v
	| Typed_var(ve) -> Typed_var(ve) (* Unused by hypothesis *)
	| Static_class(te) -> Static_class(te) (* Unused by hypothesis *) 
	| Static_interface(te) -> 
	    Static_interface(te)(* Unused by hypothesis *)
	| Assign_name(qid,s,e1) -> 
	    Assign_name(qid,s,rename exc_type old new_e e1)
	| Assign_local_var(lve,s,e1) -> (* Unused by hypothesis *) 
	    Assign_local_var(lve,s,
			     rename exc_type old new_e e1) (* Ouch! Unsafe *)
	| Assign_field(fa,s,e1) -> 
	    Assign_field(fa,s,rename exc_type old new_e e1)
	| Assign_array(e1,e2,s,e3) -> 
	    Assign_array(rename exc_type old new_e e1,
			 rename exc_type old new_e e2,
			 s,
			 rename exc_type old new_e e3)
	| Typed_assign_array(jt,e1,e2,s,e3) -> (* Unused by hypothesis *) 
	    Typed_assign_array(jt,
                               rename exc_type old new_e e1,
                               rename exc_type old new_e e2,
                               s,
                               rename exc_type old new_e e3)
	| If_expr(e1,e2,e3) -> If_expr(rename exc_type old new_e e1, 
                                       rename exc_type old new_e e2,
                                       rename exc_type old new_e e3)
	| This -> This
	| Field_access(fe) -> Field_access(fe)
	| Method_call(e1,id,e_list) -> 
	    Method_call(Option.map (rename exc_type old new_e) e1,
			id,
			List.map (rename exc_type old new_e)
			  e_list)
	| Typed_method_call _ -> assert false
	| Class_instance_creation(qi,e_list) -> 
	    Class_instance_creation(qi, 
				    List.map (rename exc_type old new_e) 
				      e_list)
	| Typed_class_instance_creation(te,e_list) -> 
	    (*i Unused by hypothesis i*) 
	    Typed_class_instance_creation(te,
					  List.map (rename exc_type old new_e)
					    e_list)
	| Array_creation(ac) as a -> a
	| Array_access(e1,e2) -> Array_access(rename exc_type old new_e e1,
                                              rename exc_type old new_e e2)
	| Typed_array_access(jt,e1,e2) -> (* Unused by hypothesis *) 
	    Typed_array_access(jt,
                               rename exc_type old new_e e1,
                               rename exc_type old new_e e2)
	| Cast(te,e1) -> Cast(te,rename exc_type old new_e e1)
	| Typed_cast(t,e1) -> assert false
	    (*i Typed_cast(te,rename exc_type old new_e e1) i*)
	| Instanceof(e1,te) -> Instanceof(rename exc_type old new_e e1,te)
	| Typed_instanceof(e1,jt) -> (* Unused by hypothesis *) 
	    Typed_instanceof(rename exc_type old new_e e1,jt)
	      (* in JML annotations only *)
	| Result -> Result  
	| Old(e1) -> Old(rename exc_type old new_e e1)
	| Fresh(e1) -> Fresh(rename exc_type old new_e e1)
	| Quantifier(q,te,vid_list,e1,e2) ->
	    Quantifier(q,te,vid_list, (* BEWARE *)
                       rename exc_type old new_e e1,
                       rename exc_type old new_e e2)
	| Typed_quantifier(q,lve,e1,e2) -> (* Unused by hypothesis *) 
	    Typed_quantifier(q,lve, 
			     rename exc_type old new_e e1,
			     rename exc_type old new_e e2); }
;;


(* [krakexception] is the identifier of the generic exception.
   Due to Why choices, its name is [result].  *)
	  
let krakexception = (Location.dummy, "result")
;;

(*i INCORRECT BECAUSE OF POSSIBLE VARIABLES CLASHES i*)

(* [mk_compact_signal_clause l] returns a single signal clause with a
   conjunction of all signal clauses in list [l]. The resulting clause
   is an matching of [Exception result] on exception types in [l].
   Note that a clause stating that the declared exception are the only
   to be thrown is added to this conjunction. *)

let mk_compact_signal_clause l =
  let (_,new_s) as new_id = krakexception in
    match l with
      | [] -> { signals_exception = type_of_Exception;
		signals_variable  = Some(new_id);
		signals_post = expr_true; }
      | clause::r ->
	  { signals_exception = type_of_Exception;
	    signals_variable  = Some(new_id);
	    signals_post = 
	      let instanceof_clause = 
		mk_expr Location.dummy
		  (Instanceof(mk_expr Location.dummy
				(Var(new_id)), 
				clause.signals_exception)) in
	      let (acc_exclu, acc_impl) =
		List.fold_left
		  (fun (acc_e,acc_i) sc -> 
		     let instanceof_expr = 
		       mk_expr Location.dummy
			 (Instanceof(mk_expr Location.dummy
				       (Var(new_id)), 
				       sc.signals_exception)) in
		       (mk_expr Location.dummy 
			  (Bin(acc_e, "||", instanceof_expr)),
			  mk_expr Location.dummy 
			    (Bin(acc_i, "&&",
				 mk_expr Location.dummy
				   (Bin(instanceof_expr, "==>",
					match sc.signals_variable with
					  | None -> sc.signals_post
					  | Some(loc,var_s) -> 
					      rename sc.signals_exception 
						(*i type_of_Exception i*)
						var_s
						new_s
						sc.signals_post))))))
		  (instanceof_clause,
		   mk_expr Location.dummy 
		     (Bin(instanceof_clause, "==>",
			  match clause.signals_variable with
			    | None -> clause.signals_post
			    | Some(loc,var_s) -> 
				rename clause.signals_exception 
				  (*i type_of_Exception i*)
				  var_s
				  new_s
				  clause.signals_post)))
		  r in
		(mk_expr Location.dummy (Bin(acc_exclu,"&&",acc_impl))) ; } 
;;   

(*i 	      mk_expr Location.dummy 
		(List.fold_left
		   (fun acc s -> mk_expr Location.dummy 
			(Bin(acc,"&&", mk_expr Location.dummy
			       (Bin( mk_expr Location.dummy
				       (Instanceof( mk_expr Location.dummy
						      (Var(new_id)), 
						      s.signals_exception)), "==>",
				       match s.signals_variable with
					 | None -> s.signals_post
					 | Some(loc,var_s) -> 
					     rename type_of_Exception
                                               var_s
					       new_s
					       s.signals_post)))))
		   (mk_expr Location.dummy 
		      (Bin( (List.fold_left 
				(fun acc_exclu sc -> 
				   mk_expr Location.dummy 
				     (Bin(mk_expr Location.dummy
					    (Instanceof (mk_expr Location.dummy
							   (Var(new_id)), 
							   sc.signals_exception)), "||", acc_exclu)))
				( mk_expr Location.dummy 
				    (Instanceof (mk_expr Location.dummy
						   (Var(new_id)), 
						   clause.signals_exception))) r), "&&",
		       (mk_expr Location.dummy 
			  (Bin(mk_expr Location.dummy
				(Instanceof(mk_expr Location.dummy
					      (Var(new_id)),
					      clause.signals_exception)), "==>",
				match clause.signals_variable with
				  | None -> clause.signals_post
				  | Some(loc,var_clause) -> 
				      rename type_of_Exception
					var_clause
					new_s
					clause.signals_post)))
		       ))) r).expr_node } ;;
i*)



let make_default_assignable pure =
  if pure then Assignable_nothing else Assignable_everything

let assign_list_to_assign l p =
  let flat = List.flatten l in
  match flat with
    | [] -> [(make_default_assignable p,None)]
    | _ -> flat
;;


(*i-----------------------------------------------------------i*)

let desugar s purity =
  match s with
  | Non_extending scl ->
      begin
	match scl with
	  | [] -> 
	      Compact({ compact_decls = {forall_var_decls = [];
					let_var_decls = [];};
			compact_requires =   expr_true;
			compact_when =       expr_true;
			compact_measured =   expr_true;
			compact_assignable = 
			  [ (make_default_assignable purity, None) ];
			compact_ensures = expr_true;
			compact_signals =
			  mk_compact_signal_clause [];
(*i
			  { signals_exception = type_of_Exception ;
			    signals_variable = Some(krakexception) ;
			    signals_post = expr_true }; 
i*)
			compact_diverges = expr_true;
		      }
		   )
	  | [Behavior(privacy,gen_spec_case)] ->
	      let body = gen_spec_case.generic_spec_body in
		begin
		  match body with
		    | Simple_spec_body(simple) ->
			Compact({ compact_decls = 
				    gen_spec_case.generic_spec_var_decls;
				  compact_requires = (*i  i*) 
 expr_list_to_conj 
   gen_spec_case.generic_spec_header.requires_clauses;
				  compact_when     =  (*i  i*) 
 expr_list_to_conj 
   gen_spec_case.generic_spec_header.when_clauses;
(*i ? i*)
				  compact_measured =  (*i  i*) 
 expr_list_to_conj 
   gen_spec_case.generic_spec_header.measured_clauses;

				  compact_assignable = 
				    assign_list_to_assign 
				      simple.assignable_clauses purity ;
				  compact_ensures =
				    expr_list_to_conj simple.ensures_clauses ;
				  compact_signals =
				    mk_compact_signal_clause 
				      simple.signals_clauses ;
				  compact_diverges =
				    expr_list_to_conj simple.diverges_clauses })
		    | _ -> desugaring_failure "desugar(1)"	
	    end
	| [Normal_behavior(privacy,normal_spec_case)] -> 
	    let body = normal_spec_case.normal_spec_body in
	      begin
		match body with 
		  | Simple_normal_spec_body(normal) ->
			Compact({ compact_decls =
				    normal_spec_case.normal_spec_var_decls;
				  compact_requires =  (*i i*) 
 expr_list_to_conj 
   normal_spec_case.normal_spec_header.requires_clauses;
				  compact_when     =   (*i i*) 
 expr_list_to_conj 
   normal_spec_case.normal_spec_header.when_clauses;
(*i ? i*)
				  compact_measured =   (*i i*) 
 expr_list_to_conj 
   normal_spec_case.normal_spec_header.measured_clauses;

				  compact_assignable = 
				    assign_list_to_assign
				      normal.normal_assignable_clauses purity;
				  compact_ensures =
				    expr_list_to_conj 
				      normal.normal_ensures_clauses ;
				  compact_signals =
				    { signals_exception = type_of_Exception ;
				      signals_variable = Some(krakexception) ;
				      signals_post = expr_false; };
				  compact_diverges =
				    expr_list_to_conj 
				      normal.normal_diverges_clauses })
		  | _ -> desugaring_failure "desugar(1.1)"
	      end
	| [Exceptional_behavior(privacy,exceptional_spec_case)] -> 
	    let body = exceptional_spec_case.exceptional_spec_body in
	      begin
		match body with 
		  | Simple_exceptional_spec_body(exceptional) ->
		      Compact({ compact_decls = 
				  exceptional_spec_case.exceptional_spec_var_decls;
			       compact_requires =  (*i i*) 
 expr_list_to_conj 
   exceptional_spec_case.exceptional_spec_header.requires_clauses;
			       compact_when     =  (*i i*) 
 expr_list_to_conj 
   exceptional_spec_case.exceptional_spec_header.when_clauses;
			       (*i ? i*)
			       compact_measured =  (*i i*) 
 expr_list_to_conj 
   exceptional_spec_case.exceptional_spec_header.measured_clauses;
			       
			       compact_assignable = 
				  assign_list_to_assign
				    exceptional.exceptional_assignable_clauses
				    purity;
			       compact_ensures = expr_false;
			       compact_signals =
				 mk_compact_signal_clause 
				   exceptional.exceptional_signals_clauses;
			       compact_diverges =
				 expr_list_to_conj 
				   exceptional.exceptional_diverges_clauses; })
		  | _ -> desugaring_failure "desugar(1.2)"
	      end
	| _ -> desugaring_failure ("desugar(2: List.length = " ^ (string_of_int (List.length scl)) ^ ")")
      end
  | Extending_also scl -> desugaring_failure "'also' unsupported"
  | Extending_and csl -> desugaring_failure "'and' unsupported"
  | Compact _ -> assert false
  | Typed_compact _ -> assert false


let desugar_assignable_clauses clauses =
  match clauses with
    | None -> None
    | Some [] -> assert false
    | Some clauses -> Some (List.flatten clauses)


let desugar_statement_spec spec =
  {
    compact_decls = { forall_var_decls = []; let_var_decls = [] };
    compact_requires = expr_list_to_conj spec.statement_requires;
    compact_when = expr_true;
    compact_measured = expr_zero;
    compact_diverges = expr_true;
    compact_assignable =
     assign_list_to_assign spec.statement_assignable false;
    compact_ensures = expr_list_to_conj spec.statement_ensures;
    compact_signals = 
     match spec.statement_signals with
       | [] -> 
	    { signals_exception = type_of_Exception ;
	      signals_variable = Some(krakexception) ;
	      signals_post = expr_false; };
       | _ ->
	   mk_compact_signal_clause spec.statement_signals
  }
